"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from pysces.version import __version__

__doc__ = '''
          PyscesLink
          ----------

          Interfaces to external software and API's, has replaced the PySCeS contrib classes.
          '''

try:
    input = raw_input  # Py2 compatibility
except NameError:
    pass

# for METATOOLlink
import os, re, io

# for SBWWebLink
try:
    from urllib.request import (
        ProxyHandler,
        build_opener,
        HTTPHandler,
        install_opener,
        urlopen,
    )  # Py 3
except ImportError:
    from urllib2 import (
        ProxyHandler,
        build_opener,
        HTTPHandler,
        install_opener,
        urlopen,
    )  # Py 2

try:
    from urllib.parse import urlencode  # Py 3
except ImportError:
    from urllib import urlencode  # Py 2


class SBWlink(object):
    """Generic access for local SBW services using SBWPython """

    sbw = None
    psbw = None
    sbwModuleProxy = None
    moduleDict = None
    modules = None

    def __init__(self):
        try:
            import SBW as SBW
            import SBW.psbw as psbw

            ##  reload(SBW)
            ##  reload(psbw)
            self.sbw = SBW
            self.psbw = SBW.psbw
            self.sbwModuleProxy = SBW.sbwModuleProxy
            self.moduleDict = SBW.sbwModuleProxy.moduleDict
            self.modules = []
            for m in self.moduleDict:
                if self.moduleDict[m].pythonName not in ['python']:
                    self.SBW_exposeAll(self.moduleDict[m])
                    self.modules.append(self.moduleDict[m].pythonName)
                    setattr(self, self.moduleDict[m].pythonName, self.moduleDict[m])
            print('\nSBWlink established.')
        except Exception as ex:
            print(ex)
            print('\nSBWlink not established.')

    def SBW_exposeAll(self, module):
        for s in module.services:
            s = getattr(module, s)
            for m in s.methods:
                getattr(s, m)

    def SBW_getActiveModules(self):
        idlist = []
        namelst = []
        for id in self.psbw.getModuleIdList():
            idlist.append(id)
            namelst.append(self.psbw.getModuleName(id))
        for id in list(self.moduleDict.keys()):
            if id not in idlist:
                self.moduleDict.pop(id)
        for name in range(len(self.modules) - 1, -1, -1):
            if self.modules[name] not in namelst:
                delattr(self, self.modules[name])
                self.modules.pop(name)
        for name in namelst:
            if name not in self.modules:
                self.SBW_loadModule(name)
        return namelst

    def SBW_loadModule(self, module_name):
        ans = 'Y'
        if module_name[-3:] == 'GUI':
            ans = input(
                'Warning! This may hang the console\n\yPress \'Y\' to continue: '
            )
        if ans == 'Y':
            module_id = self.psbw.SBWGetModuleInstance(module_name)
            assert module_id != None, '\nUnknow module, %s' % module_name
            module = self.sbwModuleProxy.ModuleProxy(module_id)
            self.SBW_exposeAll(module)
            if module_id not in self.moduleDict:
                print(
                    '<PySCeS_SBW> Adding '
                    + module.pythonName
                    + ' to ModuleProxy (id='
                    + str(module_id)
                    + ')'
                )
                self.moduleDict.update({module_id: module})
            if module.pythonName not in self.modules:
                print('<PySCeS_SBW> Adding ' + module.pythonName + ' to SBWlink')
                self.modules.append(module.pythonName)
                setattr(self, module.pythonName, module)
        else:
            print('\nModule %s not loaded' % module_name)


class SBWLayoutWebLink(object):
    """Enables access to DrawNetwork and SBMLLayout web services at www.sys-bio.org"""

    sbwhost = '128.208.17.26'
    sbml = None
    sbmllayout = None
    svg = None
    DEBUGMODE = False
    DEBUGLEVEL = 1
    DRAWNETWORKLOADED = False
    LAYOUTMODULELOADED = False

    def setProxy(self, **kwargs):
        """Set as many proxy settings as you need. You may supply a user name without
        a password in which case you will be prompted to enter one (once) when required
        (NO guarantees, implied or otherwise, on password security AT ALL). Arguments can be:

        user = 'daUser',
        pwd = 'daPassword',
        host = 'proxy.paranoid.net',
        port = 3128
        """
        proxy_info = {}
        for k in list(kwargs.keys()):
            proxy_info.update({k: kwargs[k]})

        if 'user' in proxy_info and 'pwd' not in proxy_info:
            proxy_info.update({'pwd': getpass.getpass()})
        proxy_support = ProxyHandler(
            {"http": "http://%(user)s:%(pwd)s@%(host)s:%(port)d" % proxy_info}
        )
        opener = build_opener(proxy_support, HTTPHandler)
        install_opener(opener)
        del proxy_info, proxy_support

    def loadSBMLFileFromDisk(self, File, Dir=None):
        if Dir != None:
            path = os.path.join(Dir, File)
        else:
            path = File
        if os.path.exists(path):
            self.sbmllayout = None
            self.svg = None
            self.DRAWNETWORKLOADED = False
            self.LAYOUTMODULELOADED = False
            sbmlF = open(path, 'r')
            self.sbml = sbmlF.read()
            sbmlF.close()
            return True
        else:
            print("%s is an invalid path" % path)
            return False

    def loadSBMLFromString(self, str):
        self.sbmllayout = None
        self.svg = None
        self.DRAWNETWORKLOADED = False
        self.LAYOUTMODULELOADED = False
        self.sbml = str
        return True

    def urlGET(self, host, urlpath):
        url = 'http://%s%s' % (host, urlpath)
        con = urlopen(url)
        resp = con.read()
        if self.DEBUGMODE:
            print(con.headers)
        if self.DEBUGMODE and self.DEBUGLEVEL == 2:
            print(resp)
        con.close()
        return resp

    def urlPOST(self, host, urlpath, data):
        assert type(data) == dict, '\nData must be a dictionary'
        url = 'http://%s%s' % (host, urlpath)
        con = urlopen(url, urlencode(data))
        resp = con.read()
        if self.DEBUGMODE:
            print(con.headers)
        if self.DEBUGMODE and self.DEBUGLEVEL == 2:
            print(resp)
        con.close()
        return resp

    def getVersion(self):
        print('Inspector.getVersion()')
        ver = self.urlGET(self.sbwhost, '/generate/Inspector.asmx/getVersion')
        ver = ver.replace('<?xml version="1.0" encoding="utf-8"?>', '')
        ver = ver.replace('<string xmlns="http://www.sys-bio.org/">', '')
        ver = ver.replace('</string>', '')
        return ver

    def drawNetworkLoadSBML(self):
        print('DrawNetwork.loadSBML()')
        assert self.sbml != None, '\nNo SBML file loaded'
        data = {'var0': self.sbml}
        self.DRAWNETWORKLOADED = True
        return self.urlPOST(self.sbwhost, '/generate/DrawNetwork.asmx/loadSBML', data)

    def drawNetworkGetSBMLwithLayout(self):
        print('DrawNetwork.getSBML()')
        assert self.DRAWNETWORKLOADED, '\nSBML not loaded into DrawNetwork module'
        sbml = self.urlGET(self.sbwhost, '/generate/DrawNetwork.asmx/getSBML')
        sbml = sbml.replace('&gt;', '>')
        sbml = sbml.replace('&lt;', '<')
        sbml = sbml.replace(
            '''<string xmlns="http://www.sys-bio.org/"><?xml version="1.0" encoding="utf-8"?>''',
            '',
        )
        sbml = sbml.replace('</string>', '')
        self.sbmllayout = sbml
        return True

    def layoutModuleLoadSBML(self):
        print('SBMLLayoutModule.loadSBML()')
        assert self.sbmllayout != None, '\nNo SBML Layout loaded'
        data = {'var0': self.sbmllayout}
        self.LAYOUTMODULELOADED = True
        return self.urlPOST(
            self.sbwhost, '/generate/SBMLLayoutModule.asmx/loadSBML', data
        )

    def layoutModuleGetSVG(self):
        assert self.LAYOUTMODULELOADED, '\nSBML not loaded into SBMLLayout module'
        svg = self.urlGET(self.sbwhost, '/generate/SBMLLayoutModule.asmx/getSVG')
        svg = svg.replace('&gt;', '>')
        svg = svg.replace('&lt;', '<')
        svg = svg.replace('''<string xmlns="http://www.sys-bio.org/">''', '')
        svg = svg.replace('''<?xml version="1.0" encoding="utf-8"?>''', '')
        svg = svg.replace('</string>', '')
        self.svg = svg
        return True

    def getSBML(self):
        return self.sbml

    def getSBMLlayout(self):
        return self.sbmllayout

    def getSVG(self):
        return self.svg


class METATOOLlink(object):
    """New interface to METATOOL binaries"""

    __metatool_path__ = None
    __mod__ = None
    __emode_exe_int__ = None
    __emode_exe_dbl__ = None
    __emode_intmode__ = 0
    __emode_userout__ = 0
    __emode_file__ = None
    __metatool_file__ = None
    # EModes = ''

    def __init__(self, mod, __metatool_path__=None):
        # Initialise elementary modes
        self.__mod__ = mod
        if __metatool_path__ == None:
            self.__metatool_path__ = os.path.join(mod.__pysces_directory__, 'metatool')
        else:
            self.__metatool_path__ = os.path.join(__metatool_path__, 'metatool')
        assert self.__metatool_path__ != None, '\nPySCeS not found'

        self.__emode_file__ = self.__mod__.ModelFile[:-4] + '_emodes'
        self.__metatool_file__ = self.__mod__.ModelFile[:-4] + '_metatool'
        if os.sys.platform == 'win32':
            self.__emode_exe_int__ = os.path.join(
                self.__metatool_path__, 'meta43_int.exe'
            )
            self.__emode_exe_dbl__ = os.path.join(
                self.__metatool_path__, 'meta43_double.exe'
            )
        else:
            self.__emode_exe_int__ = os.path.join(self.__metatool_path__, 'meta43_int')
            self.__emode_exe_dbl__ = os.path.join(
                self.__metatool_path__, 'meta43_double'
            )

        if os.path.exists(self.__emode_exe_int__):
            print('Using METATOOL int', end=' ')
            self.__emode_intmode__ = True
        else:
            self.__emode_exe_int__ = None
        if os.path.exists(self.__emode_exe_dbl__):
            print('\b\b\b\bdbl')
            self.__emode_intmode__ = False
        else:
            self.__emode_exe_dbl__ = None
        assert (
            self.__emode_exe_dbl__ != None or self.__emode_exe_int__ != None
        ), "\nMETATOOL binaries not available"

    def doEModes(self):
        """
        doEModes()

        Calculate the elementary modes by way of an interface to MetaTool.

        METATOOL is a C program developed from 1998 to 2000 by Thomas Pfeiffer (Berlin)
        in cooperation with Stefan Schuster and Ferdinand Moldenhauer (Berlin) and Juan Carlos Nuno (Madrid).
        http://www.biologie.hu-berlin.de/biophysics/Theory/tpfeiffer/metatool.html

        Arguments:
        None

        """
        print(
            'METATOOL is a C program developed from 1998 to 2000 by Thomas Pfeiffer (Berlin)'
        )
        print(
            'in cooperation with Stefan Schuster and Ferdinand Moldenhauer (Berlin) and Juan Carlos Nuno (Madrid).'
        )
        print(
            'http://www.biologie.hu-berlin.de/biophysics/Theory/tpfeiffer/metatool.html'
        )

        goMode = 0
        fileIn = 'pysces_metatool.dat'
        fileOut = 'pysces_metatool.out'

        goMode = 1
        if goMode == 1:
            # Build MetaTool input file
            File = open(os.path.join(self.__mod__.ModelOutput, fileIn), 'w')
            # Determine type of reaction
            out1 = []
            for key in self.__mod__.__nDict__:
                # print key
                # print self.__mod__.__nDict__[key]['Type']
                out1.append((key, self.__mod__.__nDict__[key]['Type']))
            # print '\nExtracting metatool information from network dictionary ...\n'
            File.write('-ENZREV\n')
            for x in out1:
                if x[1] == 'Rever':
                    File.write(x[0] + ' ')
            File.write('\n\n')
            File.write('-ENZIRREV\n')
            for x in out1:
                if x[1] == 'Irrev':
                    File.write(x[0] + ' ')
            File.write('\n\n')
            File.write('-METINT\n')
            for x in self.__mod__.__species__:
                File.write(x + ' ')
            File.write('\n\n')
            File.write('-METEXT\n')
            for x in self.__mod__.__fixed_species__:
                File.write(x + ' ')
            File.write('\n\n')

            output = []
            allInt = 1
            for x in self.__mod__.__nDict__:
                reList = self.__mod__.__nDict__[x]['Reagents']
                subs = ''
                prods = ''
                # print 'Reaction: ' + x
                for y in reList:
                    if self.__emode_intmode__ == 1:  # use int elementary modes
                        if abs(int(reList[y])) / abs(float(reList[y])) != 1.0:
                            print('INFO: Coefficient not integer = ' + repr(reList[y]))
                            allInt = 0
                        if reList[y] < 0:
                            # print y.replace('self.','') + ' : substrate'
                            if abs(int(reList[y])) != 1:
                                subs += repr(abs(int(reList[y]))) + ' '
                            subs += y.replace('self.', '')
                            subs += ' + '
                        else:
                            # print y.replace('self.','') + ' : product '
                            if abs(int(reList[y])) != 1:
                                prods += repr(abs(int(reList[y]))) + ' '
                            prods += y.replace('self.', '')
                            prods += ' + '
                        # output.append(x + ' : ' + subs[:-3] + ' = ' + prods[:-3] + ' .')
                    else:  # use float/double elementary mode
                        if reList[y] < 0.0:
                            # print y.replace('self.','') + ' : substrate'
                            if abs(float(reList[y])) != 1.0:
                                subs += repr(abs(float(reList[y]))) + ' '
                            subs += y.replace('self.', '')
                            subs += ' + '
                        else:
                            # print y.replace('self.','') + ' : product '
                            if abs(float(reList[y])) != 1.0:
                                prods += repr(abs(float(reList[y]))) + ' '
                            prods += y.replace('self.', '')
                            prods += ' + '
                output.append(x + ' : ' + subs[:-3] + ' = ' + prods[:-3] + ' .')

            File.write('-CAT\n')
            for x in output:
                File.write(x + '\n')
            File.write('\n')

            File.flush()
            File.close()

            if allInt == 1:
                if self.__emode_intmode__ == 1:
                    eModeExe = self.__emode_exe_int__
                else:
                    eModeExe = self.__emode_exe_dbl__

                print('\nMetatool running ...\n')
                ######### UPDATE:
                # Actually works fine on windows and posix - johann 20081128
                print('Generic run')
                os.spawnl(
                    os.P_WAIT,
                    eModeExe,
                    eModeExe,
                    os.path.join(self.__mod__.ModelOutput, fileIn),
                    os.path.join(self.__mod__.ModelOutput, fileOut),
                )
                print('\nMetatool analysis complete\n')

                # Parse MetaTool output file and store the result in a string
                go = 0
                go2 = 0
                result = ''
                end = ''

                try:
                    file2 = open(os.path.join(self.__mod__.ModelOutput, fileOut), 'r')
                    for line in file2:
                        c = re.match('ELEMENTARY MODES', line)
                        d = re.match(' enzymes', line)
                        e = re.match('The elementary mode', line)
                        f = re.match('\n', line)
                        g = re.match('The elementary', line)
                        if c != None:
                            go = 1
                            go2 = 0
                        if d != None:
                            go2 = 1
                        if e != None:
                            go2 = 0
                        if go == 1 and go2 == 1 and f == None:
                            line = line.replace('reversible', '\n  reversible\n')
                            line = line.replace('ir\n  ', '\n  ir')
                            if self.__emode_intmode__ == 1:
                                line = line.replace('] ', ']\n ')
                            else:
                                line = line.replace(') ', ')\n ', 1)
                            result += line
                        if go == 1 and g != None:
                            end += line
                    result += end
                    result += '\n'

                    file2.close()

                    if self.__emode_userout__ == 1:
                        fileo = open(
                            os.path.join(
                                self.__mod__.ModelOutput, self.__metatool_file__
                            )
                            + '.in',
                            'w',
                        )
                        filer = open(
                            os.path.join(self.__mod__.ModelOutput, fileIn), 'r'
                        )
                        for line in filer:
                            fileo.write(line)
                        fileo.write('\n\n')
                        filer.close()
                        fileo.close()
                        filer = open(
                            os.path.join(self.__mod__.ModelOutput, fileOut), 'r'
                        )
                        fileo = open(
                            os.path.join(
                                self.__mod__.ModelOutput, self.__metatool_file__
                            )
                            + '.out',
                            'w',
                        )
                        for line in filer:
                            fileo.write(line)
                        filer.close()
                        fileo.close()
                    os.remove(os.path.join(self.__mod__.ModelOutput, fileIn))
                    os.remove(os.path.join(self.__mod__.ModelOutput, fileOut))
                except Exception as EX:
                    print('doEmode:', EX)
                    print(
                        'WARNING: Unable to open MetaTool output file\nPlease check the MetaTool executables: '
                    )
                    if os.name == 'posix':
                        print(
                            '/MetaTool/meta43_double /MetaTool/meta43_int\nand their permissions'
                        )
                    else:
                        print('/MetaTool/meta43_double.exe /MetaTool/meta43_int.exe')
            else:
                print(
                    '\nINFO: non-integer coefficients\
                \nTry using the double eMode function: self.__emode_intmode__=0'
                )
                result = 'Elementary modes not calculated\n'
        else:
            print('\nNo elementary mode calculation possible - no meta43_xxx.exe')
            result = 'Elementary modes not calculated\n'
        self.EModes = result

    def getEModes(self):
        """
        getEModes()

        Returns the elementary modes as a linked list of fluxes

        """
        try:
            a = self.EModes
            FF = io.StringIO()
            FF.write(self.EModes)
            FF.reset()
            output = []
            for line in FF:
                if (
                    re.match('  ', line)
                    and not re.match('  reversible', line)
                    and not re.match('  irreversible', line)
                ):
                    tmp = [el for el in line.replace('\n', '').split(' ') if el != '']
                    tmpOut = []
                    skip = False
                    for el in range(len(tmp)):
                        if skip:
                            skip = False
                        elif tmp[el][0] != '(':
                            tmpOut.append(tmp[el])
                        elif tmp[el][0] == '(':
                            tmpOut.append(tmp[el] + ')' + tmp[el + 1][:-1])
                            skip = True
                    output.append(tmpOut)
            return output
        except AttributeError as atx:
            print(atx)
            print('\nINFO: Please run doEModes() first\n')

    def showEModes(self, File=None):
        """
        showEModes(File=None)

        Print the results of an elementary mode analysis, generated with doEModes(),
        to screen or file.

        Arguments:
        File [default=None]: Boolean, if True write parsed elementary modes to file

        """
        try:
            if File != None:
                # assert type(File) == file, 'showEmodes() needs an open file object'
                print('\nElementary modes written to file\n')
                f = open(
                    os.path.join(
                        self.__mod__.ModelOutput, self.__emode_file__ + '.out'
                    ),
                    'w',
                )
                f.write('\n## Elementary modes\n')
                f.write(self.EModes)
                f.close()
            else:
                print('\nElementary modes\n')
                print(self.EModes)
        except AttributeError as atx:
            print(atx)
            print('\nINFO: Please run doEModes() first\n')


'''
_HAVE_STOMPY = False
_STOMPY_LOAD_ERROR = ''
try:
    ##  import stompy
    import stochpy as stompy
    _HAVE_STOMPY = True
except Exception, ex:
    _STOMPY_LOAD_ERROR = '%s' % ex
    _HAVE_STOMPY = False


class StomPyInterface(object):
    """
    StomPy interface to PySCeS this may move to pysces.link in the future
    """
    SSA = None
    SSA_REACTIONS = None
    SSA_SPECIES = None
    stompy = None
    _MOD2PSC = None

    TMP_FNAME = None
    TMP_PATH = None
    MODEL_PATH = None
    OUTPUT_PATH = None

    STP_IS_TIME_SIM = False
    STP_METHOD = 'Direct'
    STP_TIMEEND = 1
    STP_TRAJ = 1
    STP_INTERACTIVE = True
    STP_TRACK_PROPENSITIES = True
    STP_WAITING_TIMES = True
    STP_STEPS = 10
    STP_INITIAL_SPECIES = True
    STP_KEEP_PSC_FILES = False

    def __init__(self, model_path, output_path):
        """
        An interface class to the StomPy stochastic simulator

         - *model_path* the default PySCeS model directory
         - *output_path* the default PySCeS output directory
        """
        self.stompy = stompy
        self.OUTPUT_PATH = output_path
        self.MODEL_PATH = model_path
        self.TMP_PATH = os.path.join(model_path, 'orca')
        self._MOD2PSC = interface.writeMod2PSC


    def setProperty(self, **kwargs):
        """
        Sets a StomPy simulation parameter

         - *method* [default='Direct'] select simulation algorithm
         - *trajectories* [default=1]
         - *interactive* [default=True]
         - *track_propensities* [default=True]
         - *steps* [default=10]

        """
        ##  print kwargs

        if kwargs.has_key('method'):
            self.STP_METHOD = kwargs['method']
            ##  print '%s = %s' % ('method', kwargs['method'])
        if kwargs.has_key('trajectories'):
            self.STP_TRAJ = kwargs['trajectories']
            self.STP_TRAJ = 1 # TODO I need to look into this
            ##  print 'Currently only single trajectories are supported via the PySCeS interface'
            ##  print '%s = %s' % ('trajectories', self.STP_TRAJ)
        if kwargs.has_key('interactive'):
            self.STP_INTERACTIVE = kwargs['interactive']
            ##  print '%s = %s' % ('interactive', kwargs['interactive'])
        if kwargs.has_key('track_propensities'):
            self.STP_TRACK_PROPENSITIES = kwargs['track_propensities']
            ##  print '%s = %s' % ('track_propensities', kwargs['track_propensities'])
        if kwargs.has_key('steps'):
            self.STP_STEPS = kwargs['steps']
            ##  print '%s = %s' % ('steps', kwargs['steps'])
        if kwargs.has_key('species_initial'):
            self.STP_INITIAL_SPECIES = kwargs['initial_species']
            ##  print '%s = %s' % ('initial_species', kwargs['initial_species'])
        if kwargs.has_key('keep_psc_files'):
            self.STP_KEEP_PSC_FILES = kwargs['keep_psc_files']
            ##  print '%s = %s' % ('keep_psc_files', kwargs['keep_psc_files'])

    def initModelFromMod(self, pscmod, iValues=False):
        """
        Initialise a StomPy SSA instance from a PySCeS model.

         - *pscmod* an initialised PySCeS model
         - *iValues* [default=False] use initial values (not current)

        """
        self.TMP_FNAME = str(time.time()).replace('.','')+'.psc'

        if self.STP_INITIAL_SPECIES:
            for s in pscmod.species:
                setattr(pscmod, s, pscmod.__sDict__[s]['initial'])

        self._MOD2PSC(pscmod, self.TMP_FNAME, self.TMP_PATH, iValues=iValues)

        self.SSA = self.stompy.SSA(Method=self.STP_METHOD, File=self.TMP_FNAME, dir=self.TMP_PATH, IsInteractive=self.STP_INTERACTIVE)
        self.SSA.Trajectories(self.STP_TRAJ)
        self.SSA_REACTIONS = self.SSA.SSA.rate_names
        self.SSA_SPECIES = self.SSA.SSA.species

        if self.STP_TRACK_PROPENSITIES:
            self.SSA.TrackPropensities()
        try:
            print os.path.join(self.TMP_PATH, self.TMP_FNAME)
            if not self.STP_KEEP_PSC_FILES and self.TMP_PATH != None and self.TMP_FNAME != None:
                os.remove(os.path.join(self.TMP_PATH, self.TMP_FNAME))
        except:
            print 'Could not delete intermediatery StomPy PSC file: %s' % os.path.join(self.TMP_PATH, self.TMP_FNAME)
        self.TMP_FNAME = None
        print 'StomPy model ... initialised.'

    def runTimeSimulation(self, pscmod, endtime=None, method='Direct', iValues=False):
        """
        Run a stochastic simulation

         - *pscmod* and instanitiated PySCeS model
         - *endtime* [default=1] the end time **Note: this could take a long time i.e. generate ahuge amount of steps**
         - *method* [default='Direct'] select the simulation method, one of:

          - *Direct*
          - *FirstReactionMethod*
          - *NextReactionMethod*
          - *TauLeaping*

         - *iValues* [default=False] use initial values (not current)

        """

        if method not in ['Direct','FirstReactionMethod','NextReactionMethod','TauLeaping']:
            print 'Method: %s does not exist using - Direct' % method
            self.STP_METHOD = 'Direct'
        else:
            self.STP_METHOD = method
        if endtime != None:
            self.STP_TIMEEND = endtime
        self.initModelFromMod(pscmod, iValues=iValues)

        ##  self.SSA.Timesteps(self.STP_STEPS)
        self.SSA.Endtime(self.STP_TIMEEND)
        self.SSA.Run()
        self.STP_IS_TIME_SIM = True
        ##  self.SSA.PlotTimeSim()

        print 'StomPy time simulation ... done.'

        # TODO STOCHPY

        ##  if self.SSA.SSA.output[0][-1] == '':
            ##  self.SSA.SSA.output[0][-1] = 0.0
        ##  sim_dat = numpy.array(self.SSA.SSA.output, 'd')

        ##  pscmod.data_stochsim = IntegrationStochasticDataObj()
        ##  pscmod.data_stochsim.setTime(sim_dat[:,0])
        ##  pscmod.data_stochsim.setSpecies(sim_dat[:,1:-1], self.SSA_SPECIES)

        pscmod.data_stochsim = self.SSA.data_stochsim

        if self.STP_WAITING_TIMES:
            wtimes, wt_lbls = self.getWaitingtimesData(reactions=None,lbls=True)
            pscmod.data_stochsim.setWaitingtimes(wtimes, wt_lbls)
        if self.STP_TRACK_PROPENSITIES:
            pscmod.data_stochsim.setPropensities(self.SSA.SSA.propensities_output)
        pscmod.data_stochsim.TYPE_INFO = 'Stochastic'

    def runStepSimulation(self, pscmod, steps=None, method='Direct', iValues=False):
        """
        Run a stochastic simulation

         - *pscmod* and instanitiated PySCeS model
         - *steps* [default=10] the number of steps to simulate
         - *method* [default='Direct'] select the simulation method, one of:

          - *Direct*
          - *FirstReactionMethod*
          - *NextReactionMethod*
          - *TauLeaping*

        - *iValues* [default=False] use initial values (not current)

        """

        if method not in ['Direct','FirstReactionMethod','NextReactionMethod','TauLeaping']:
            print 'Method: %s does not exist using - Direct' % method
            self.STP_METHOD = 'Direct'
        else:
            self.STP_METHOD = method
        if steps != None:
            self.STP_STEPS = steps
        self.initModelFromMod(pscmod, iValues=iValues)

        self.SSA.Timesteps(self.STP_STEPS)
        ##  self.SSA.Endtime(self.STP_TIMEEND)
        self.SSA.Run()
        self.STP_IS_TIME_SIM = False

        print 'StomPy step simulation ... done.'
        ##  print self.SSA.SSA.output[0]
        ##  print self.SSA.SSA.output[1]
        ##  print self.SSA.SSA.output[-1]
        ##  header_line = self.SSA.SSA.output.pop(0)
        ##  if self.SSA.SSA.output[0][-1] == '':
            ##  self.SSA.SSA.output[0][-1] = 0.0
        ##  sim_dat = numpy.array(self.SSA.SSA.output, 'd')

        ##  pscmod.data_stochsim = IntegrationStochasticDataObj()
        ##  pscmod.data_stochsim.setTime(sim_dat[:,0])
        ##  pscmod.data_stochsim.setSpecies(sim_dat[:,1:-1], self.SSA_SPECIES)

        pscmod.data_stochsim = self.SSA.data_stochsim

        if self.STP_WAITING_TIMES:
            wtimes, wt_lbls = self.getWaitingtimesData(reactions=None,lbls=True)
            pscmod.data_stochsim.setWaitingtimes(wtimes, wt_lbls)
        if self.STP_TRACK_PROPENSITIES:
            pscmod.data_stochsim.setPropensities(self.SSA.SSA.propensities_output)
        pscmod.data_stochsim.TYPE_INFO = 'Stochastic'


    def getWaitingtimesData(self,reactions=None,lbls=False):
        """
        Plots the waiting times for each reaction in the model. Makes use of ObtainWaitingtimes to derive the waiting times out of the SSA output.

        Input:

         - *reactions* [default=0] a list of reactions to plot defualts to all reactions
         - *traj* [default=0] trajectory to plot (defaults to first one)
         - *lbls* [default=False] if True return (data_array, column_labels) otherwise just data_array

        This method is derived from StomPy 0.9 (http://stompy.sf.net) Analysis.py

        """
        if self.SSA.IsTauLeaping:
            print 'INFO: waiting times not available when method is Tau Leaping'
            if not lbls:
                return None
            else:
                return None, None
        self.SSA.GetWaitingtimes()
        if reactions == None:
            reactions = self.SSA_REACTIONS
        vect = []
        vect_lbls = []
        for r in reactions:
            if r in self.SSA_REACTIONS:
                vect.append(self.SSA_REACTIONS.index(r)+1)
                vect_lbls.append('wt'+str(r))
            else:
                print "INFO: '%s' is not a valid reaction name" % r
        OUTPUT = []
        ##  for t in range(len(self.SSA.data_stochsim.waiting_times)):
        T_OUTPUT = []
        for i in vect:
            if self.SSA.data_stochsim.waiting_times.has_key(i):
                waiting_time = self.SSA.data_stochsim.waiting_times[i]
                if len(waiting_time) > 1:			# At least 2 waiting times are necessary per reaction
                    T_OUTPUT.append(self.stompy.modules.Analysis.LogBin(waiting_time, 1.5)) 	# Create logarithmic bins
                else:
                    T_OUTPUT.append(None)
            else:
                T_OUTPUT.append(None)
        OUTPUT.append(T_OUTPUT)
        if not lbls:
            return OUTPUT
        else:
            return OUTPUT, vect_lbls


class IntegrationStochasticDataObj(object):
    """
    This class is specifically designed to store the
    results of a stochastic time simulation
    It has methods for setting the Time, Labels, Species and Propensity data and
    getting Time, Species and Rate (including time) arrays. However, of more use:

    - getOutput(\*args) feed this method species/rate labels and it will return
      an array of [time, sp1, r1, ....]
    - getDataAtTime(time) the data generated at time point "time".
    - getDataInTimeInterval(time, bounds=None) more intelligent version of the above
      returns an array of all data points where: time-bounds <= time <= time+bounds

    """
    time = None
    waiting_times = None
    species = None
    propensities = None
    xdata = None
    time_label = 'Time'
    waiting_time_labels = None
    species_labels = None
    propensities_labels = None
    xdata_labels = None
    HAS_SPECIES = False
    HAS_WAITING_TIMES = False
    HAS_PROPENSITIES = False
    HAS_TIME = False
    HAS_XDATA = False
    IS_VALID = True
    TYPE_INFO = 'Stochastic'

    def setLabels(self, species):
        """
        Set the species

         - *species* a list of species labels
        """
        self.species_labels = species

    def setTime(self, time, lbl=None):
        """
        Set the time vector

         - *time* a 1d array of time points
         - *lbl* [default=None] is "Time" set as required

        """
        self.time = time.reshape(len(time), 1)
        self.HAS_TIME = True
        if lbl != None:
            self.time_label = lbl

    def setSpecies(self, species, lbls=None):
        """
        Set the species array

         - *species* an array of species vs time data
         - *lbls* [default=None] a list of species labels
        """
        self.species = species
        self.HAS_SPECIES = True
        if lbls != None:
            self.species_labels = lbls

    def setWaitingtimes(self, waiting_times, lbls=None):
        """
        Set the `waiting_times` this data structure is not an array but a nested list of: waiting time log bins per reaction per trajectory::

        waiting_times = [traj_1, ..., traj_n]
        traj_1 = [wt_J1, ..., wt_Jn] # in order of SSA_REACTIONS
        wt_J1 = (xval, yval, nbin)
        xval =[x_1, ..., x_n]
        yval =[y_1, ..., y_n]
        nbin = n

         - *waiting_times* a list of waiting times
         - *lbls* [default=None] a list of matching reaction names

        """
        self.waiting_times = waiting_times
        self.HAS_WAITING_TIMES = True
        if lbls != None:
            self.waiting_time_labels = lbls


    def setPropensities(self, propensities, lbls=None):
        """
        Sets an array of propensities.

         - *propensities* a list of propensities
         - *lbls* [default=None] a list of matching reaction names

        """

        if lbls == None:
            LB = copy.copy(propensities[0])
            lbls = LB[1:]
            lbls = ['p'+str(r) for r in lbls]
        P_ARR = numpy.zeros((len(propensities), len(propensities[0])-1), 'd')
        P_ARR[-1,:] = numpy.NaN
        for r in range(1, P_ARR.shape[0]):
            P_ARR[r, :] = propensities[r][1:]
        self.propensities = P_ARR
        self.HAS_PROPENSITIES = True
        if lbls != None:
            self.propensities_labels = lbls
        ##  print self.propensities_labels
        ##  print self.propensities


    def setXData(self, xdata, lbls=None):
        """
        Sets an array of extra simulation data

        - *xdata* an array of xdata vs time
        - *lbls* [default=None] a list of xdata labels

        """
        self.xdata = xdata
        self.HAS_XDATA = True
        if lbls != None:
            self.xdata_labels = lbls

    def getTime(self, lbls=False):
        """
        Return the time vector

         - *lbls* [default=False] return only the time array or optionally both the time array and time label

        """
        output = None
        if self.HAS_TIME:
            output = self.time.reshape(len(self.time),)
        if not lbls:
            return output
        else:
            return output, [self.time_label]

    def getSpecies(self, lbls=False):
        """
        Return an array fo time+species

        - *lbls* [default=False] return only the time+species array or optionally both the data array and a list of column label

        """
        output = None
        if self.HAS_SPECIES:
            output = numpy.hstack((self.time, self.species))
            labels = [self.time_label]+self.species_labels
        else:
            output = self.time
            labels = [self.time_label]
        if not lbls:
            return output
        else:
            return output, labels

    def getWaitingtimes(self, lbls=False, traj=[]):
        """
        Return waiting times, time+waiting_time array

         - *lbls* [default=False] return only the time+waiting_time array or optionally both the data array and a list of column label
         - *traj* [default=[0]] return the firs or trajectories defined in this list

        """
        output = None
        labels = None
        if self.HAS_WAITING_TIMES:
            output = []
            if len(traj) == 0:
                traj = range(len(self.waiting_times))
            ##  if len(traj) == 1:
                ##  output = self.waiting_times[0]
            ##  else:
            for t in traj:
                output.append(self.waiting_times[t])
            labels = self.waiting_time_labels
        else:
            output = []
            labels = []
        if not lbls:
            return output
        else:
            return output, labels

    def getPropensities(self, lbls=False):
        """
        Return time+propensity array

         - *lbls* [default=False] return only the time+propensity array or optionally both the data array and a list of column label

        """
        #assert self.propensities != None, "\nNo propensities"
        output = None
        if self.HAS_PROPENSITIES:
            print self.time.shape
            print self.propensities.shape
            output = numpy.hstack((self.time, self.propensities))
            labels = [self.time_label]+self.propensities_labels
        else:
            output = self.time
            labels = [self.time_label]
        if not lbls:
            return output
        else:
            return output, labels

    def getXData(self, lbls=False):
        """
        Return time+xdata array

        - *lbls* [default=False] return only the time+xdata array or optionally both the data array and a list of column label

        """
        output = None
        if self.HAS_XDATA:
            output = numpy.hstack((self.time, self.xdata))
            labels = [self.time_label]+self.xdata_labels
        else:
            output = self.time
            labels = [self.time_label]
        if not lbls:
            return output
        else:
            return output, labels

    def getDataAtTime(self, time):
        """
        Return all data generated at "time"

         - *time* the required exact time point

        """
        #TODO add rate rule data
        t = None
        sp = None
        ra = None
        ru = None
        xd = None
        temp_t = self.time.reshape(len(self.time),)
        for tt in range(len(temp_t)):
            if temp_t[tt] == time:
                t = tt
                if self.HAS_SPECIES:
                    sp = self.species.take([tt], axis=0)
                if self.HAS_PROPENSITIES:
                    ru = self.propensities.take([tt], axis=0)
                if self.HAS_XDATA:
                    xd = self.xdata.take([tt], axis=0)
                break

        output = None
        if t is not None:
            output = numpy.array([[temp_t[t]]])
            if sp is not None:
                output = numpy.hstack((output,sp))
            if ra is not None:
                output = numpy.hstack((output,ra))
            if ru is not None:
                output = numpy.hstack((output,ru))
            if xd is not None:
                output = numpy.hstack((output,xd))

        return output

    def getDataInTimeInterval(self, time, bounds=None):
        """
        Returns an array of all data in interval: time-bounds <= time <= time+bounds
        where bound defaults to stepsize

         - *time* the interval midpoint
         - *bounds* [default=None] interval halfspan defaults to stepsize

        """

        temp_t = self.time.reshape(len(self.time),)
        if bounds == None:
            bounds = temp_t[1] - temp_t[0]
        c1 = (temp_t >= time-bounds)
        c2 = (temp_t <= time+bounds)
        print 'Searching (%s:%s:%s)' % (time-bounds, time, time+bounds)

        t = []
        sp = None
        ra = None

        for tt in range(len(c1)):
            if c1[tt] and c2[tt]:
                t.append(tt)
        output = None
        if len(t) > 0:
            output = self.time.take(t)
            output = output.reshape(len(output),1)
            if self.HAS_SPECIES and self.HAS_TIME:
                output = numpy.hstack((output, self.species.take(t, axis=0)))
            if self.HAS_PROPENSITIES:
                output = numpy.hstack((output, self.propensities.take(t, axis=0)))
            if self.HAS_XDATA:
                output = numpy.hstack((output, self.xdata.take(t, axis=0)))
        return output

    def getAllSimData(self,lbls=False):
        """
        Return an array of time + all available simulation data

         - *lbls* [default=False] return only the data array or (data array, list of labels)

        """
        labels = [self.time_label]
        if self.HAS_SPECIES and self.HAS_TIME:
            output = numpy.hstack((self.time, self.species))
            labels += self.species_labels
        if self.HAS_PROPENSITIES:
            output = numpy.hstack((output, self.propensities))
            labels += self.propensities_labels
        if self.HAS_XDATA:
            output = numpy.hstack((output, self.xdata))
            labels += self.xdata_labels
        if not lbls:
            return output
        else:
            return output, labels

    def getSimData(self, *args, **kwargs):
        """
        Feed this method species/xdata labels and it will return an array of [time, sp1, ....]

         - 'speces_l', 'xdatal' ...
         - *lbls* [default=False] return only the data array or (data array, list of labels)

        """
        output = self.time

        if kwargs.has_key('lbls'):
            lbls = kwargs['lbls']
        else:
            lbls = False
        lout = [self.time_label]
        for roc in args:
            if self.HAS_SPECIES and roc in self.species_labels:
                lout.append(roc)
                output = numpy.hstack((output, self.species.take([self.species_labels.index(roc)], axis=-1)))
            if self.HAS_PROPENSITIES and roc in self.propensities_labels:
                lout.append(roc)
                output = numpy.hstack((output, self.propensities.take([self.propensities_labels.index(roc)], axis=-1)))
            if self.HAS_XDATA and roc in self.xdata_labels:
                lout.append(roc)
                output = numpy.hstack((output, self.xdata.take([self.xdata_labels.index(roc)], axis=-1)))
        if not lbls:
            return output
        else:
            return output, lout


class PysMod:
#STOMPY INSERT START
    def StochSimPlot(self, plot='species', filename=None, title=None, log=None, format='points'):
        """
        Plot the Stochastic simulation results, uses the new UPI pysces.plt interface:

        - *plot* [default='species'] output to plot, can be one of:

         - 'all' species and propensities
         - 'species' species
         - 'waiting_times' waiting_times
         - 'propensities' propensities
         - `['S1', 'R1', ]` a list of model attributes ('species')

        - *filename* [default=None] if defined file is exported to filename
        - *title* [default=None] the plot title
        - *log* [default=None] use log axis for axis 'x', 'y', 'xy'
        - *format* [default='lines'] use UPI or backend specific keys
        """
        data = None
        labels = None
        allowedplots = ['all', 'species', 'propensities','waiting_times']
        ##  allowedplots = ['all', 'species', 'waiting_times']
        if type(plot) != list and plot not in allowedplots:
            raise RuntimeError, '\nPlot must be one of %s not \"%s\"' % (str(allowedplots), plot)
        if plot == 'all':
            ##  data, labels = self.data_stochsim.getSpecies(lbls=True)
            data, labels = self.data_stochsim.getAllSimData(lbls=True)
        elif plot == 'species':
            data, labels = self.data_stochsim.getSpecies(lbls=True)
        elif plot == 'propensities':
            data, labels = self.data_stochsim.getPropensities(lbls=True)
            ##  data, labels = self.data_stochsim.getRates(lbls=True)
        elif plot == 'waiting_times':
            dataLst, labels = self.data_stochsim.getWaitingtimes(lbls=True)
            format='points'
            ##  data, labels = self.data_stochsim.getRates(lbls=True)
        else:
            plot = [at for at in plot if at in self.__species__+[self.data_stochsim.time_label]+self.data_stochsim.propensities_labels]
            kwargs = {'lbls' : True}
            print plot
            if len(plot) > 0:
                data, labels = self.data_stochsim.getSimData(*plot, **kwargs)
        del allowedplots

        xu = 'Time (%(multiplier)s x %(kind)s x 10**%(scale)s)**%(exponent)s' % self.__uDict__['time']

        if plot == 'waiting_times':
            xu = 'Inter-arrival time (%s)' % xu

            xrng_start = 0.1
            xrng_end = 0.1
            yrng_start = 0.1
            yrng_end = 0.1
            for wt in range(len(dataLst)):
                for d in range(len(dataLst[wt])):
                    D = dataLst[wt][d]
                    if plt.__USE_MATPLOTLIB__ and d > 0:
                        plt.m.hold(True)
                    if D != None and len(D[0]) > 0 and len(D[1]) > 0:
                        data = numpy.vstack([D[0], D[1]]).transpose()
                        if min(D[0]) < xrng_start and min(D[0]) > 0.0:
                            xrng_start = min(D[0])
                        if max(D[0]) > xrng_end:
                            xrng_end = max(D[0])
                        if min(D[1]) < yrng_start and min(D[1]) > 0.0:
                            yrng_start = min(D[1])
                        if max(D[1]) > yrng_end:
                            yrng_end = max(D[1])
                        plt.plotLines(data, 0, [1], titles=['Time']+[labels[d]], formats=[format])
            plt.setRange('x', xrng_start*0.8, xrng_end*1.2)
            plt.setRange('y', yrng_start*0.8, yrng_end*1.2)
            if plt.__USE_MATPLOTLIB__:
                plt.m.hold(False)
        else:
            plt.plotLines(data, 0, range(1, data.shape[1]), titles=labels, formats=[format])
            # set the x-axis range so that it is original range + 0.2*sim_end
            # this is a sceintifcally dtermned amount of space that is needed for the title at the
            # end of the line :-) - brett 20040209
            RngTime = self.data_stochsim.getTime()
            end = RngTime[-1] + 0.2*RngTime[-1]
            plt.setRange('x', RngTime[0], end)
            del RngTime

        # For now StochPy results are plotted as Amounts directly from StochPy
        M = 'Amount'
        ##  if self.__KeyWords__['Output_In_Conc']:
            ##  M = 'Concentration'
        ##  else:
            ##  M = 'Amount (%(multiplier)s x %(kind)s x 10**%(scale)s)**%(exponent)s' % self.__uDict__['substance']

        if plot == 'all':
            yl = 'Amount, propensities'
        elif plot == 'propensities':
            yl = 'Propensities'
        elif plot == 'waiting_times':
            yl = 'Frequency'
            if log == None:
                log = 'xy'
        elif plot == 'species':
            yl = '%s' % M
        else:
            yl = 'User defined'
        plt.setAxisLabel('x', xu)
        plt.setAxisLabel('y', yl)
        if log != None:
            plt.setLogScale(log)
        if title == None:
            plt.setGraphTitle('PySCeS/StochPy simulation (' + self.ModelFile + ') ' + time.strftime("%a, %d %b %Y %H:%M:%S"))
        else:
            plt.setGraphTitle(title)
        plt.replot()
        if filename != None:
            plt.export(filename, directory=self.ModelOutput, type='png')


    def doStochSim(self,end=10,mode='steps',method='Direct',trajectories=1):
        """
        doStochSim(end=10, mode='steps', method='Direct')

        Run a stochastic simulation for until `end` is reached. This can be either steps or end time (which could be a *HUGE* number of steps).

        Arguments:

         - *end* [default=10] simulation end (steps or time)
         - *mode* [default='steps'] simulation mode, can be one of:

          - *steps* total number of steps to simulate
          - *time* simulate until time is reached

         - *method* [default='Direct'] stochastic algorithm, can be one of:

          - Direct
          - FirstReactionMethod
          - NextReactionMethod
          - TauLeaping

        """

        if method not in ['Direct', 'FirstReactionMethod','NextReactionMethod','TauLeaping']:
            print 'Method "%s" not recognised using: "Direct"' % method
            method = 'Direct'
        if mode not in ['steps','time']:
            print 'Mode "%s" not recognised using: "steps"' % mode
            mode = 'steps'

        stompy_track_propensities = True
        stompy_keep_psc_files = False
        self.__STOMPY__.setProperty(method=method, trajectories=trajectories, interactive=True, track_propensities=stompy_track_propensities, keep_psc_files=stompy_keep_psc_files)

        if mode == 'time':
            self.__STOMPY__.runTimeSimulation(self, endtime=end, method=method)
        else:
            self.__STOMPY__.runStepSimulation(self, steps=end, method=method)


    def doStochSimPlot(self, end=10.0, mode='steps', method='Direct', plot='species', fmt='points', log=None, filename=None):
        """
        doStochSimPlot(end=10.0, mode='steps', method='Direct', plot='species', fmt='points', log=None, filename=None)

        Run a stochastic simulation for until `end` is reached and plot the results. This can be either steps or end time (which could be a *HUGE* number of steps).

        Arguments:

         - *end* [default=10] simulation end (steps or time)
         - *mode* [default='steps'] simulation mode, can be one of:

          - *steps* total number of 'steps' to simulate
          - *time* simulate until 'time' is reached

         - *method* [default='Direct'] stochastic algorithm, can be one of:

          - Direct
          - FirstReactionMethod
          - NextReactionMethod
          - TauLeaping

        - *plot* [default='species'] output to plot, can be one of:

         - 'all' species and propensities
         - 'species' species
         - 'waiting_times' waiting_times
         - 'propensities' propensities
         - `['S1', 'R1', ]` a list of model attributes ('species')

        - *filename* [default=None] if defined file is exported to filename
        - *title* [default=None] the plot title
        - *log* [default=None] use log axis for axis 'x', 'y', 'xy'
        - *fmt* [default='lines'] use UPI or backend specific keys

        """

        self.doStochSim(end=end, mode=mode, method=method,trajectories=1)
        self.StochSimPlot(plot='species', filename=filename, log=log, format=fmt)

#STOMPY INSERT START

if not _HAVE_STOMPY:
    def nofunc(self, *args, **kwargs):
        print '\nStochastic simulation not available, please download/install *StomPy* from: http://stompy.sf.net\n'
    PysMod.doStochSim = nofunc
    PysMod.doStochSimPlot = nofunc
    PysMod.StochSimPlot = nofunc

'''



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from .version import __version__

import os, copy, re, time
from getpass import getuser
from xml.etree import ElementTree
import io
import itertools
import numpy

from .. import output_dir, __CHGDIR_ON_START__
if __CHGDIR_ON_START__:
    CWD = output_dir
else:
    CWD = os.getcwd()

from .InfixParser import MyInfixParser

InfixParser = MyInfixParser()
InfixParser.buildlexer()
InfixParser.buildparser(
    debug=0,
    debugfile='infix.dbg',
    tabmodule='infix_tabmodule',
    outputdir=output_dir,
)
InfixParser.setNameStr('self.', '()')

os.chdir(CWD)

class CoreToPsc(object):
    core = None
    name = ''
    header_block = ''
    fixed_block = ''
    compartment_block = ''
    function_block = ''
    reaction_block = ''
    species_block = ''
    parameter_block = ''
    rrule_block = ''
    arule_block = ''
    event_block = ''
    notes_block = ''
    SPECIES_CURRENT_VALUE = False
    FIXED_SPECIES_CURRENT_VALUE = False
    PARAMETER_CURRENT_VALUE = False
    time = None
    pw_symbols = None
    __DEBUG__ = False

    # future
    compartment_block = ''

    def __init__(self, core):
        self.core = core
        self.name = self.core.name.replace('.psc', '')
        self.name = self.core.name.replace('.xml', '')
        self.pw_symbols = [pw.name for pw in self.core.piecewise_functions]

    def setHeader(self, s=''):
        out = '# Generated by PySCeS %s (%s)\n' % (
            __version__,
            time.strftime("20%y-%m-%d %H:%M"),
        )
        if len(s) > 0:
            out += '\n# '
        ccount = 0
        for c in range(len(s)):
            ccount += 1
            if ccount >= 60 and ccount <= 75 and s[c] in [' ', '\t', '\n', '-']:
                out += s[c] + '\n# '
                ccount = 0
            else:
                out += s[c]
        out += ' \n# Keywords\n'
        kk = list(self.core.__KeyWords__.keys())
        kk.sort()
        for key in kk:
            if self.core.__KeyWords__[key] != None:
                if key == 'ModelType':
                    pass  # TODO
                    ##  out += '%s: ' % (key)
                    ##  for mt in range(len(self.core.__KeyWords__[key])):
                    ##  out += '%s,' % (self.core.__KeyWords__[key][mt])
                    ##  out = out[:-1] + '\n'
                else:
                    out += '%s: %s\n' % (key, self.core.__KeyWords__[key])
        out += ' \n# GlobalUnitDefinitions\n'
        for key in list(self.core.__uDict__.keys()):
            if key in ['substance', 'volume', 'time', 'length', 'area']:
                k1, k2 = key[0].upper(), key[1:]
                k = k1 + k2
                v = self.core.__uDict__[key]
                out += 'Unit%s: %s, %s, %s, %s\n' % (
                    k,
                    v['kind'],
                    v['multiplier'],
                    v['scale'],
                    v['exponent'],
                )
        out += ' \n'
        self.header_block = out

    def setFixed(self):
        out = ''
        fcnt = 0
        for f in self.core.species:
            if f.fixed:
                fcnt += 1
                out += '%s ' % f.name
        if fcnt > 0:
            out = 'FIX: ' + out
            out += '\n \n'
        self.fixed_block = out

    def setCompartments(self):
        out = ''
        if len(self.core.compartments) > 0:
            out += '# Compartments\n'
            for c in self.core.compartments:
                ra = ''
                if c.area != None:
                    ra = ', ' + c.area
                out += 'Compartment: %s, %s, %s %s\n' % (
                    c.name,
                    c.size,
                    c.dimensions,
                    ra,
                )
            out += ' \n'
        self.compartment_block = out

    def setFunctions(self):
        out = ''
        start = True
        for f in self.core.functions:
            if start:
                out = '# Function definitions\n'
                start = False
            out += 'Function: %s' % f.name
            for a in f.args:
                out += ', %s ' % a
            out += ' {\n%s\n}\n' % f.formula
        if out != '':
            out += ' \n'
        self.function_block = out

    def setReactions(self):
        out = '# Reactions'
        for r in self.core.reactions:
            C = ''
            if not r.hasCompartment() and len(self.core.compartments) == 1:
                print(
                    'Info: single compartment model: locating \"%s\" in default compartment'
                    % r.name
                )
                C = '@%s' % self.core.compartments[0].getName()
            elif not r.hasCompartment() and len(self.core.compartments) > 1:
                print(
                    'Info: multiple compartments defined but reaction %s is not located in any of them.'
                    % r.name
                )
            elif r.hasCompartment():
                C = '@%s' % r.getCompartment().getName()

            rout = '\n%s%s:\n    ' % (r.name, C)
            cnt = 0

            # we assume that multistoich_enabled will only be true if NewCore.netStoich is False
            ##  print r.hasSubstrates()
            ##  print r.hasProducts()

            if not r.multistoich_enabled:
                for lh in r.hasSubstrates():
                    ##  print '\nFuck'
                    ##  print lh
                    ##  print r.stoichiometry
                    if cnt != 0:
                        rout += ' + '
                    if abs(r.stoichiometry[lh]) == 1.0:
                        rout += '%s' % lh
                    else:
                        rout += '{%s}%s' % (abs(r.stoichiometry[lh]), lh)
                    cnt += 1
                if len(r.hasSubstrates()) == 0:
                    rout += '$pool'
                if r.reversible:
                    rout += ' = '
                else:
                    rout += ' > '
                cnt = 0

                for rh in r.hasProducts():
                    ##  print '\nFuck2'
                    ##  print rh
                    ##  print r.stoichiometry
                    if cnt != 0:
                        rout += ' + '
                    if abs(r.stoichiometry[rh]) == 1.0:
                        rout += '%s' % rh
                    else:
                        rout += '{%s}%s' % (abs(r.stoichiometry[rh]), rh)
                    cnt += 1
                if len(r.hasProducts()) == 0:
                    rout += '$pool'
            else:
                LHS = ''
                RHS = ''
                REVS = ''
                for lh in r.multistoich:
                    if lh[1] < 0.0:
                        if abs(lh[1]) == 1.0:
                            LHS += '%s' % lh[0]
                        else:
                            LHS += '{%s}%s' % ((lh[1]), lh[0])
                        LHS += ' + '
                    else:
                        if abs(lh[1]) == 1.0:
                            RHS += '%s' % lh[0]
                        else:
                            RHS += '{%s}%s' % (abs(lh[1]), lh[0])
                        RHS += ' + '

                if LHS == '':
                    LHS += '$pool'
                else:
                    LHS = LHS[:-3]
                if RHS == '':
                    RHS += '$pool'
                else:
                    RHS = RHS[:-3]
                if r.reversible:
                    REVS += ' = '
                else:
                    REVS += ' > '
                rout = '%s%s %s %s' % (rout, LHS, REVS, RHS)

            # if we have piecewise symbols test if they are in the formula
            # and replace them with the piecewise expression
            formula = r.formula
            if len(self.pw_symbols) > 0:
                pwreplace = 'piecewise('
                for pwd in self.pw_symbols:
                    if pwd in r.formula:
                        for pwc in self.core.__piecewises__[pwd]:
                            if pwc != 'other':
                                pwreplace += '%s,%s,' % (
                                    self.core.__piecewises__[pwd][pwc][0],
                                    self.core.__piecewises__[pwd][pwc][1],
                                )
                        if self.core.__piecewises__[pwd]['other'] != None:
                            pwreplace += '%s) ' % self.core.__piecewises__[pwd]['other']
                        else:
                            pwreplace = pwreplace[:-1] + ')'
                        formula = formula.replace(pwd, pwreplace)
            rout += '\n    %s\n' % formula
            if len(r.hasModifiers()) > 0:
                rout += '# %s has modifier(s): ' % r.name
                for m in r.hasModifiers():
                    rout += '%s ' % m
                rout += ' \n'
            out += rout
        out += ' \n'
        self.reaction_block = out

    def setSpecies(self):
        out = ''
        fixed = ''
        var = ''
        for f in self.core.species:
            C = ''
            if not f.hasCompartment() and len(self.core.compartments) == 1:
                print(
                    'Info: single compartment model %s is in default compartment'
                    % f.name
                )
                C = '@%s' % self.core.compartments[0].getName()
            elif not f.hasCompartment() and len(self.core.compartments) > 1:
                print(
                    'Warning: multiple compartments defined but species %s is not located in any of them.'
                    % f.name
                )
            elif f.hasCompartment():
                C = '@%s' % f.getCompartment().getName()

            if not f.fixed:
                if not self.SPECIES_CURRENT_VALUE:
                    value = f.value_initial
                else:
                    value = f()
                var += '%s%s = %s\n' % (f.name, C, value)
            else:
                if not self.FIXED_SPECIES_CURRENT_VALUE:
                    value = f.value_initial
                else:
                    value = f()
                fixed += '%s%s = %s\n' % (f.name, C, value)
        out += '# Fixed species\n'
        out += fixed
        out += ' \n'
        out += '# Variable species\n'
        out += var
        out += ' \n'
        self.species_block = out

    def setParameters(self):
        out = '# Parameters\n'
        processed = []

        def setP(p2, out2, proced):
            if not self.PARAMETER_CURRENT_VALUE:
                # take into account that parameters can be compartments now as well
                # brett 2008
                if hasattr(p2, 'value_initial') and p2.value_initial != None:
                    value = p2.value_initial
                else:
                    value = p2()
            else:
                value = p2()
            if p2.name in proced:
                # out2 += '# '
                pass
            else:
                out2 += '%s = %s' % (p2.name, value)
            if not hasattr(p2, 'code_string'):
                if p2.name in proced:
                    pass
                    # out2 += '\t# (initialised)'
                else:
                    out2 += '\n'
            else:
                if p2.name in proced:
                    pass
                    # out2 += '\t# (initialised)\n'
                else:
                    out2 += '\t# (rule)\n'
            return out2

        for r in self.core.reactions:
            ##  out += '# %s parameters\n' % r.name
            for p in r.parameters:
                if p.name not in self.core.hasCompartments():
                    out = setP(p, out, processed)
                    processed.append(p.name)
            ##  out += '\n'
        ##  if len(processed) != len(self.core.hasGlobalParameters()):
        ##  out += '# Additional (global) parameters\n'
        for p in self.core.hasGlobalParameters():
            if p not in processed:
                p = self.core.get(p)
                out = setP(p, out, processed)
                processed.append(p.name)
        out += ' \n'
        self.parameter_block = out

    def setAssignmentRules(self):
        out = ''
        start = True
        rids = [r.name for r in self.core.reactions]
        for p in self.core.global_parameters + self.core.species:
            if hasattr(p, 'type') and getattr(p, 'type') == 'assignment':
                if start:
                    out = '# Assignment rules\n'
                    start = False
                # add support for reaction names, symbol repalce R --> R()
                formula = p.formula
                for n_ in p._names:
                    if n_ in rids:
                        formula = formula.replace(n_, '{}()'.format(n_))

                # if we have piecewise symbols test if they are in the formula
                # and replace them with the piecewise expression
                if len(self.pw_symbols) > 0:
                    pwreplace = ' piecewise('
                    for pwd in self.pw_symbols:
                        if pwd in p.formula:
                            for pwc in self.core.__piecewises__[pwd]:
                                if pwc != 'other':
                                    pwreplace += '%s,%s,' % (
                                        self.core.__piecewises__[pwd][pwc][0],
                                        self.core.__piecewises__[pwd][pwc][1],
                                    )
                            if self.core.__piecewises__[pwd]['other'] != None:
                                pwreplace += (
                                    '%s) ' % self.core.__piecewises__[pwd]['other']
                                )
                            else:
                                pwreplace = pwreplace[:-1] + ')'
                            formula = formula.replace(pwd, pwreplace)
                out += '!F %s = %s\n' % (p.name, formula)
        if out != '':
            out += ' \n'
        self.arule_block = out

    def setRateRules(self):
        out = ''
        start = True
        if len(self.core.rate_rules) > 0:
            for rr in self.core.rate_rules:
                if start:
                    out = '# Rate rules\n'
                    start = False
                out += 'RateRule: %s {\n%s\n}\n' % (rr.name, rr.formula)
            out += '\n'
        self.rrule_block = out

    def setEvents(self):
        out = ''
        start = True
        for ev in self.core.events:
            if start:
                out = '# Event definitions\n'
                start = False
            ##  formula = ev.code_string.split('=',1)
            formula = ev.formula
            out += 'Event: %s, %s, %s \n{\n' % (ev.name, formula, ev.delay)
            for ass in ev.assignments:
                out += '%s = %s\n' % (ass.variable.name, ass.formula)
            out += '}\n'
        if out != '':
            out += ' \n'
        self.event_block = out

    def setNotes(self, s=''):
        if len(s) > 1:
            out = '# Notes\n# '
            out += '"""%s\n"""\n' % s
            ##  ccount = 0
            ##  for c in range(len(s)):
            ##  ccount += 1
            ##  if ccount >= 55 and ccount <=75 and s[c] in [' ', '\t', '\n', '-']:
            ##  out += s[c]+'\n# '
            ##  ccount = 0
            ##  else:
            ##  out += s[c]
            ##  out += ' \n'
            self.notes_block = out
        else:
            self.notes_block = '\n'

    def write(self, filename=None, directory=None, getstrbuf=False):
        if filename == None:
            filename = self.name + '.psc'
        if directory != None:
            assert os.path.exists(directory), '\n%s does not exist.' % directory
            filename = os.path.join(directory, filename)
        print('Writing file: %s' % filename)
        outF = open(filename, 'w')
        outF.write(self.header_block)
        outF.write(self.fixed_block)
        outF.write(self.compartment_block)
        outF.write(self.function_block)
        outF.write(self.reaction_block)
        outF.write(self.rrule_block)
        outF.write(self.arule_block)
        outF.write(self.event_block)
        outF.write(self.species_block)
        outF.write(self.parameter_block)
        outF.write(self.notes_block)
        outF.flush()
        outF.close()
        if getstrbuf:
            fb = io.StringIO()
            outF = open(filename, 'r')
            fb.write(outF.read())
            outF.flush()
            outF.close()
            return fb


class PscToCore(object):
    ModelDir = None
    WorkDir = None

    # old
    ModelFile = None
    _PysMod__InitStrings = None
    InitParams = None  # suspect
    fixed_species = None
    species = None
    parameters = None
    reactions = None
    modifiers = None
    ##  __function_forced_str__ = None
    _Function_time = None
    _Function_user = None
    _Function_init = None
    __nDict__ = None

    __functions__ = None
    __rules__ = None
    time = None
    __events__ = None
    __uDict__ = None

    def __init__(self):
        try:
            from pysces.PyscesParse import PySCeSParser

            self.pscParser = PySCeSParser(debug=0)
        except:
            print("\nYou need PySCeS installed to use this module")

    def getPSCFileFromDisk(self, ModelFile, ModelDir=None, WorkDir=None):
        """
        find and set up an existing psc file for parsing
        """

        if ModelFile[-4:] != '.psc':
            print("Assuming .psc extension")
            ModelFile += '.psc'
        if ModelDir == None:
            # this will probably change in future - bgoli
            ModelDir = CWD
        if WorkDir == None:
            WorkDir = CWD
        assert os.path.exists(
            os.path.join(ModelDir, ModelFile)
        ), '\nFile %s does not exist' % os.path.join(ModelDir, ModelFile)
        setattr(self, 'ModelFile', ModelFile)
        setattr(self, 'ModelDir', ModelDir)
        assert os.path.exists(WorkDir), '\nDirectory %s does not exist' % WorkDir
        setattr(self, 'WorkDir', WorkDir)

    def getPSCFileFromString(self, ModelString, WorkDir=None):
        """
        set up a temporary psc file for parsing from model string
        """
        if WorkDir == None:
            WorkDir = CWD
        assert os.path.exists(WorkDir), '\nDirectory %s does not exist' % WorkDir
        setattr(self, 'WorkDir', WorkDir)
        setattr(self, 'ModelDir', WorkDir)
        ModelFile = '%s.psc' % time.time()
        Mfile = open(os.path.join(WorkDir, ModelFile), 'w')
        Mfile.write(ModelString)
        Mfile.close()
        setattr(self, 'ModelFile', ModelFile)

    def getParsedModel(self):
        """
        Parse the input file associated with the PySCeS model instance and assign the basic model attributes
        """

        self.pscParser.ParsePSC(self.ModelFile, self.ModelDir, self.WorkDir)

        # from __nDict__
        self._PysMod__InitStrings = [
            s.replace('self.', '') for s in self.pscParser.InitStrings
        ]
        for s in self.pscParser.InitStrings:
            exec(s)
        self.InitParams = [s.replace('self.', '') for s in self.pscParser.InitParams]
        self.fixed_species = copy.copy(self.pscParser.fixed_species)
        self.species = copy.copy(self.pscParser.species)
        self.parameters = copy.copy(self.pscParser.parameters)
        self.reactions = copy.copy(self.pscParser.reactions)
        self.modifiers = copy.copy(self.pscParser.modifiers)
        self.__functions__ = copy.copy(self.pscParser.functions)

        # print self.modifiers
        self._Function_forced = self.pscParser.ForceFunc.replace('self.', '')
        self._Function_forced = self.pscParser.ForceFunc.replace('scipy.', '')
        self._Function_time = self.pscParser.TimeFunc.replace('self.', '')
        self._Function_user = self.pscParser.UserFunc.replace('self.', '')
        self._Function_init = self.pscParser.InitFunc.replace('self.', '')
        self.__nDict__ = self.pscParser.NetworkDict.copy()
        self.__sDict__ = self.pscParser.sDict.copy()
        self.__compartments__ = self.pscParser.Compartments.copy()
        self.__uDict__ = self.pscParser.uDict.copy()
        # new

    def getModel(self):
        return self


# TODO: PIECEWISE WRITING


class CoreToSBML(object):
    core = None
    name = None
    SBML = None
    level = 2
    version = 1
    model = None
    document = None
    time = None
    __events__ = None
    __DEBUG__ = False
    UNINIT_DEFAULT = 1e-3

    NumpyToMathML = {
        'numpy.greater_equal': 'geq',
        'numpy.greater': 'gt',
        'numpy.less_equal': 'leq',
        'numpy.less': 'lt',
        'numpy.not_equal': 'neq',
        'numpy.equal': 'eq',
        'numpy.arcsinh': 'arcsinh',
        'numpy.arccosh': 'arccosh',
        'numpy.arctanh': 'arctanh',
        'numpy.arcsin': 'arcsin',
        'numpy.arccos': 'arccos',
        'numpy.arctan': 'arctan',
        'numpy.sinh': 'sinh',
        'numpy.cosh': 'cosh',
        'numpy.tanh': 'tanh',
        'numpy.sin': 'sin',
        'numpy.cos': 'cos',
        'numpy.tan': 'tan',
        'math.log10': 'log',
        'math.log': 'ln',
        'numpy.floor': 'floor',
        'numpy.ceil': 'ceil',
        'numpy.sqrt': 'sqrt',
        'math.sqrt': 'sqrt',
        'math.exp': 'exp',
        'operator.eq': 'eq',
        'operator.ne': 'neq',
        'operator.gt': 'gt',
        'operator.ge': 'geq',
        'operator.lt': 'lt',
        'operator.le': 'leq',
        'operator.eq': 'equal',
        'operator.ne': 'neq',
        'self._piecewise_': 'piecewise',
        '_piecewise_': 'piecewise',
        'operator.not_': 'not',
        ' and ' : ' && ',
        ' or ' : ' || ',

    }

    NumpyToMathMLkeys = [
        'numpy.greater_equal',
        'numpy.greater',
        'numpy.less_equal',
        'numpy.less',
        'numpy.not_equal',
        'numpy.equal',
        'numpy.arcsinh',
        'numpy.arccosh',
        'numpy.arctanh',
        'numpy.arcsin',
        'numpy.arccos',
        'numpy.arctan',
        'numpy.sinh',
        'numpy.cosh',
        'numpy.tanh',
        'numpy.sin',
        'numpy.cos',
        'numpy.tan',
        'math.log10',
        'math.log',
        'numpy.floor',
        'numpy.ceil',
        'numpy.sqrt',
        'math.sqrt',
        'math.exp',
        'operator.eq',
        'operator.ne',
        'operator.gt',
        'operator.ge',
        'operator.lt',
        'operator.le',
        'operator.eq',
        'operator.ne',
        'self._piecewise_',
        '_piecewise_',
        'operator.not_',
        ' and ',
        ' or ',
    ]

    def __init__(self, core):
        self.core = core
        self.name = self.core.getName().replace('.psc', '').replace('.xml', '')
        try:
            import libsbml as SBML

            self.SBML = SBML
        except Exception as e:
            print(e)
            print('Posix sbml load error')
            self.SBML = None

    def createModel(self):
        """
        Create an SBML model and document uses the class attributes:

         - *self.level* [default=2] SBML level
         - *self.version* [default=4] SBML version

        and creates:

         - self.model and SBML model
         - self.document an SBML document

        """
        if self.SBML.getLibSBMLVersion() < 40000:
            self.model = self.SBML.Model()
            self.model.setName(self.name)
            self.document = self.SBML.SBMLDocument()
            self.document.setLevelAndVersion(self.level, self.version)
        else:
            self.model = self.SBML.Model(self.level, self.version)
            self.model.setName(self.name)
            self.document = self.SBML.SBMLDocument(self.level, self.version)

    def setCompartments(self):
        if len(self.core.compartments) < 1:
            print('Warning: no compartments defined adding one called \"Cell\"')
            self.core.addOneCompartment('Cell', 1.0, 3)
        for cs in self.core.compartments:
            comp_def = self.model.createCompartment()
            comp_def.setId(cs.name)
            comp_def.setName(cs.name)
            comp_def.setVolume(float(cs.size))
            # TODO need to adapt this for rate rules
            if self.level == 3:
                comp_def.setConstant(False)

    def setDescription(self, txt=None):
        '''
        Sets the model description as a <note> containing `txt` in an HTML paragraph on the model object.

         - *txt* [default=None] the text to insert as the description

        '''

        if txt == None:
            txt = self.core.getDescription()
        else:
            self.core.setDescription(txt)
        notes = ''
        notes = '<body xhtml="http://www.w3.org/1999/xhtml>'
        notes += (
            '<p><span style="font-family: Courier New,Courier,monospace;">%s</span></p>\n'
            % txt
        )
        notes += (
            '<p><span style="font-family: Courier New,Courier,monospace;"><a href="http://pysces.sourceforge.net">PySCeS</a> (%s) generated model (%s)</span></p>'
            % (__version__, self.core.getName())
        )
        notes += '</body>'
        self.model.setNotes(notes)

    # TODO: update this to handle composite units
    def setUnits(self):
        """
        Adds the unit definitions to the SBML model object

        """
        ud = self.core.getGlobalUnits()
        for un in list(ud.keys()):
            # for now just the basic stuff
            if un in ['substance', 'volume', 'time', 'length', 'area']:
                vdef = self.model.createUnitDefinition()
                vdef.setId(un)
                vdef.setName(un)
                vu = self.model.createUnit()
                vu.setKind(self.SBML.UnitKind_forName(ud[un]['kind']))
                vu.setMultiplier(ud[un]['multiplier'])
                vu.setScale(int(ud[un]['scale']))
                vu.setExponent(int(ud[un]['exponent']))
                vu.setOffset(0)

        # new super cool way of doing arbitrary unit definitions
        """
        ud = {'area': {0 : {'exponent': 2, 'kind': 'metre', 'multiplier': 1.0, 'scale': 0}},
                    'length': {0 : {'exponent': 1, 'kind': 'metre', 'multiplier': 1.0, 'scale': 0}},
                    'substance': {0 : {'exponent': 1, 'kind': 'mole', 'multiplier': 1.0, 'scale': 0}},
                    'time': {0 : {'exponent': 1, 'kind': 'second', 'multiplier': 1.0, 'scale': 0}},
                    'volume': {0 : {'exponent': 1, 'kind': 'litre', 'multiplier': 1.0, 'scale': 0}},
                    'mmol_per_gDW_per_hr': {0 : {'exponent': 1, 'kind': 'mole', 'multiplier': 1.0, 'scale': -3},
                                            1 : {'exponent': -1, 'kind': 'gram', 'multiplier': 1.0, 'scale': 0},
                                            2 : {'exponent': -1, 'kind': 'second', 'multiplier': 0.00027777, 'scale': 0}
                                          }
                    }

        for un in ud.keys():
            vdef = model.createUnitDefinition()
            vdef.setId(un)
            vdef.setName(un)
            for u in range(len(ud[un].keys())):
                vu = model.createUnit()
                vu.setKind(libsbml.UnitKind_forName(ud[un][u]['kind']))
                vu.setMultiplier(ud[un][u]['multiplier'])
                vu.setScale(int(ud[un][u]['scale']))
                vu.setExponent(int(ud[un][u]['exponent']))
                vu.setOffset(0)
        """

    def setSpecies(self):
        """
        Add the species definitions to the SBML model object:

        """
        for spe in self.core.species:
            s = self.model.createSpecies()
            s.setId(spe.name)
            s.setName(spe.name)
            if not spe.hasCompartment() and len(self.core.compartments) == 1:
                print(
                    'Warning: species %s does not have a compartment locating it in \"%s\"'
                    % (spe.getName(), self.core.compartments[0].getName())
                )
                spe.setCompartment(self.core.compartments[0])
            elif not spe.hasCompartment() and len(self.core.compartments) >= 1:
                raise UserWarning(
                    "%s does not have a compartment and there are more than 1 to choose from!"
                    % spe.getName()
                )
            s.setCompartment(spe.getCompartment().getName())
            if spe.name in self.core.hasFixedSpecies():
                s.setBoundaryCondition(True)
                s.setConstant(True)
            else:
                s.setBoundaryCondition(False)
            if spe() is None:
                print(
                    'Warning, species {} has not been initialised setting to {}'.format(spe.getName(), self.UNINIT_DEFAULT)
                )

                if spe.name in self.core.__InitDict__ and self.core.__InitDict__[spe.name] is None:
                    self.core.__InitDict__[spe.name] = self.UNINIT_DEFAULT

                spe.value = spe.value_initial = self.core.__InitDict__[spe.name]


            if spe.isAmount():
                s.setInitialAmount(spe())
                s.setHasOnlySubstanceUnits(True)
            else:
                s.setInitialConcentration(spe())
                s.setHasOnlySubstanceUnits(False)

    def setParameters(self):
        for par in self.core.global_parameters:
            p = self.model.createParameter()
            p.setId(par.name)
            p.setName(par.name)

            # check for None value with no value and assign a default
            #print('PAR', par.name, par())
            # TODO: sort out this ZeroDivision and parameter initialisation error in a more sophisticated way, for example, evaluation assignment rules on export.
            try:
                if par() is None:
                    print(
                        'INFO: parameter \"{}\" not been defined setting to {}'.format(par.getName(), self.UNINIT_DEFAULT)
                    )
                    if par.name in self.core.__InitDict__ and self.core.__InitDict__[par.name] is None:
                        self.core.__InitDict__[par.name] = self.UNINIT_DEFAULT
                    par.value = par.value_initial = self.core.__InitDict__[par.name]
                # check for uninitialise parameters that have a value
                elif par.name in self.core.__InitDict__:
                    print(
                        'INFO: parameter \"{}\" not initialised setting to ({})'.format(par.name, par())
                    )
                    self.core.__InitDict__[par.name] = par()
                    par.value = par.value_initial = self.core.__InitDict__[par.name]
            except ZeroDivisionError:
                print('INFO: parameter \"{}\"initialisation error setting to 0.0'.format(par.name))
                par.value = par.value_initial = self.core.__InitDict__[par.name] = 0.0



            # first attempt, check for a formula ... could be done with introspection
            if hasattr(par, 'formula'):
                p.setConstant(False)
                r = self.model.createAssignmentRule()
                r.setVariable(par.name)
                formula = (
                    par.code_string.split('=')[1].replace('self.', '').replace('()', '')
                )
                #print('LOOKATME PARSE ASSIGNMENT RULE line 861')
                #print(formula)
                formula = self.infixPSC2SBML(formula)
                #print(formula)
                sbml_ast = self.SBML.parseL3Formula(formula)
                assert sbml_ast is not None, '\nERROR: cannot interpret assignment rule ({}): {}'.format(par.name, formula)
                #print(self.SBML.formulaToL3String(sbml_ast))

                if r.setMath(sbml_ast) < 0:
                    print('\nERROR: cannot assign assignment rule ({}): {}'.format(par.name, formula))
                    raise(RuntimeError)

                #r.setFormula(par.formula)
                #print(self.core.__InitDict__)
                #print(par.name)
                #print(self.core.__InitDict__[par.name])
                p.setValue(self.core.__InitDict__[par.name])
            else:
                p.setValue(par())

            # TODO need to adapt this for rate rules
            if self.level == 3:
                p.setConstant(False)


    def infixPSC2SBML(self, infix):
        """replace NumPy infix with libSBMl infix"""

        # symbol replacement parse (sometimes simple is better)
        infix2 = str(infix)
        for f in self.NumpyToMathMLkeys:
            infix2 = infix2.replace(f, self.NumpyToMathML[f])
        return infix2

    def astSetCSymbolTime(self, ast):
        """set ASTNode type <cn> _TIME_ </cn> to <csymbol> time </csymbol>"""
        strBuf = io.StringIO()
        mathMLin = self.SBML.writeMathMLToString(ast)
        strBuf.write(mathMLin)
        strBuf.seek(0)
        etree = ElementTree.parse(strBuf)
        root = etree.getroot()
        counter = itertools.count(1)

        def idxNode(node, idx=0):
            if node.text != None and node.text.strip() == '_TIME_':
                # <csymbol encoding="text" definitionURL="http://www.sbml.org/sbml/symbols/time"> t </csymbol>
                node.text = node.text.replace('_TIME_', 'time')
                node.tag = "{http://www.sbml.org/sbml/symbols/time}csymbol"
                node.attrib.update({'encoding': 'text'})
                node.attrib.update(
                    {'definitionURL': 'http://www.sbml.org/sbml/symbols/time'}
                )

            children = node.getchildren()
            for child in range(len(children)):
                idxNode(children[child], next(counter))

        idxNode(root, idx=0)

        strBuf = io.StringIO()
        etree.write(strBuf)
        strBuf.seek(0)
        mathMLout = strBuf.read()
        return self.SBML.readMathMLFromString(mathMLout)

    def setRules(self):
        """Set rate rules"""
        for rule in self.core.rate_rules:
            RR = self.model.createRateRule()
            RR.setVariable(rule.getName())
            # replace PySCeS infix with libSBML infix
            form = rule.code_string.split('=')[1].replace('self.', '').replace('()', '')
            form = self.infixPSC2SBML(form)
            req_replacements = {}
            for symb in rule._names:
                if symb in self.core.hasReactions():
                    req_replacements.update(
                        {
                            symb: '(%s)'
                            % self.infixPSC2SBML(
                                self.core.get(symb)
                                .code_string.split('=')[1]
                                .replace('self.', '')
                                .replace('()', '')
                            )
                        }
                    )
            if len(list(req_replacements.keys())) > 0:
                InfixParser.setNameStr('', '')
                InfixParser.SymbolReplacements = req_replacements
                InfixParser.parse(form)
                form = InfixParser.output
            ASTnode = self.SBML.parseFormula(form)
            assert ASTnode != None, "ERROR: unable to parse formula (%s) to AST" % form
            if self.__DEBUG__:
                print('Adding RateRule: %s = %s' % (rule.getName(), form))
            # set _TIME_ ASTnode tag to <csymbol> time </csymbol>
            ASTnode = self.astSetCSymbolTime(ASTnode)
            RR.setMath(ASTnode)
            # after creating the rule set the parameter to be non constant
            rpar = self.model.getParameter(rule.getName())
            if rpar == None:
                rpar = self.model.getCompartment(rule.getName())
            if rpar == None:
                rpar = self.model.getSpecies(rule.getName())
            if rpar != None:
                rpar.setConstant(False)

    def setFunctions(self):
        for func in self.core.functions:
            #print(func)
            FNC = self.model.createFunctionDefinition()
            FNC.setName(func.getName())
            FNC.setId(func.getName())
            form = func.code_string.split('=')[1].replace('self.', '').replace('()', '')
            #print(form)
            form = self.infixPSC2SBML(form)
            #print(form)
            ASTnode = self.SBML.parseL3Formula(form)

            assert ASTnode is not None, "ERROR: unable to parse function (%s) to AST" % form
            assert FNC.setMath(ASTnode) == 0, "ERROR: unable to set Function math."



    def setEvents(self):
        """Set events"""
        for ev in self.core.events:
            if self.__DEBUG__:
                print('Adding event: %s' % ev.getName())
            EV = self.model.createEvent()
            EV.setName(ev.getName())
            EV.setId(ev.getName())

            #print(ev.getName())
            #print(ev.formula)
            #print(ev.code_string)
            # replace PySCeS infix with libSBML infix
            #print('LOOKATME PARSE EVENT line 979')
            form = ev.code_string.replace('self.state=', '').replace('self.', '').replace('()', '')
            #print(form)
            form = self.infixPSC2SBML(form)
            #print(form)
            if self.__DEBUG__:
                print('\tTrigger: %s' % form)
            ASTnode = self.SBML.parseL3Formula(form)
            assert ASTnode != None, "ERROR: unable to parse formula (%s) to AST" % form
            #print(self.SBML.formulaToL3String(ASTnode))
            ## set _TIME_ ASTnode tag to <csymbol> time
            #ASTnode = self.astSetCSymbolTime(ASTnode)

            tr = self.SBML.Trigger(self.level, self.version)
            tr.setMath(ASTnode)
            EV.setTrigger(tr)
            ea_cntr = 0
            for ass in ev.assignments:
                #print('LOOKATME PARSE EVENTASSIGNMENT line 1003')
                if self.__DEBUG__:
                    print('\tAssignment: %s = %s' % (ass.getName(), ass.formula))
                #print(ass.getName())
                #print(ass.formula)
                #print(ass.variable.getName())
                form = self.infixPSC2SBML(ass.formula)
                #print(form)
                ASTnode = self.SBML.parseL3Formula(form)

                # TODO in future we need to move all id's from SBML to CORE including triggers and event assignments
                # this will affect this code

                eass = self.SBML.EventAssignment(self.level, self.version)
                eass.setId('eas_{}_{}'.format(ev.getName(), ea_cntr))
                ea_cntr += 1
                eass.setVariable(ass.variable.getName())
                assert eass.setMath(ASTnode) == 0, 'ERROR: EventAssignment math set error.'
                EV.addEventAssignment(eass)
            if ev.delay != 0:
                dform = self.infixPSC2SBML(ev.delay)
                ASTnodeD = self.SBML.parseFormula(dform)
                ASTnodeD = self.astSetCSymbolTime(ASTnodeD)
                D = self.SBML.Delay(ASTnodeD)
                EV.setDelay(D)

    def setReactions(self):
        for rxn in self.core.reactions:
            # print 'Adding reaction:', rxn.name
            SBML_R = self.model.createReaction()
            SBML_R.setId(rxn.name)
            SBML_R.setName(rxn.name)
            for s in rxn.substrates:
                ##  print '\t' + rxn.name +' has substrate: ' + s.name + ' (%s)' % abs(rxn.stoichiometry[s.name])
                if self.SBML.getLibSBMLVersion() < 40000:
                    sref = self.SBML.SpeciesReference(
                        s.name, abs(rxn.stoichiometry[s.name])
                    )
                else:
                    sref = self.SBML.SpeciesReference(self.level, self.version)
                    sref.setStoichiometry(abs(rxn.stoichiometry[s.name]))
                    sref.setSpecies(s.name)
                # TODO: check if this needs to be refined
                if self.level == 3:
                    sref.setConstant(True)
                SBML_R.addReactant(sref)
            for p in rxn.products:
                ##  print '\t' + rxn.name +' has product: ' + p.name + ' (%s)' % abs(rxn.stoichiometry[p.name])
                if self.SBML.getLibSBMLVersion() < 40000:
                    pref = self.SBML.SpeciesReference(
                        p.name, abs(rxn.stoichiometry[p.name])
                    )
                else:
                    pref = self.SBML.SpeciesReference(self.level, self.version)
                    pref.setStoichiometry(abs(rxn.stoichiometry[p.name]))
                    pref.setSpecies(p.name)

                # TODO: check if this needs to be refined
                if self.level == 3:
                    pref.setConstant(True)
                # print(pref.toSBML())
                SBML_R.addProduct(pref)
            if not rxn.multistoich_enabled:
                for m in rxn.modifiers:
                    self.model.createModifier().setSpecies(m.name)
            else:
                AllReagents = [r[0] for r in rxn.multistoich]
                for m in rxn.modifiers:
                    if m.name not in AllReagents:
                        # print '\t' + rxn.name +' has modifier: ' + m.name
                        self.model.createModifier().setSpecies(m.name)
            formula = (
                rxn.code_string.split('=')[1].replace('self.', '').replace('()', '')
            )
            formula = self.infixPSC2SBML(formula)
            if self.SBML.getLibSBMLVersion() < 40000:
                sb_kl = self.SBML.KineticLaw(formula)
            else:
                sb_kl = self.SBML.KineticLaw(self.level, self.version)
                sb_kl.setFormula(formula)
            SBML_R.setKineticLaw(sb_kl)

            if rxn.reversible:
                SBML_R.setReversible(True)
            else:
                SBML_R.setReversible(False)

            # TODO: check if this needs to be refined
            if self.level == 3:
                SBML_R.setFast(False)
            # print(SBML_R.toSBML())

    def getSBMLmodel(self):
        return self.model

    def getSBMLdocument(self):
        self.document.setModel(self.model)
        return self.document

    def getSBML(self):
        return self.getSBMLdocument().toSBML()

    def getSBMLFileAsStrBuf(self):
        try:
            UseR = getuser()
        except:
            UseR = ''
        fb = io.StringIO()
        h1 = '<?xml version="1.0" encoding="utf-8"?>\n'
        h1 += (
            '<!-- Created with PySCeS ('
            + __version__
            + ') on '
            + time.strftime("%a, %d %b %Y %H:%M:%S")
            + ' by '
            + UseR
            + ' -->\n'
        )
        fb.write(h1 + self.getSBML())
        return fb

    def writeSBML2file(self, filename=None, directory=None):
        """
        Writes the SBML model to file:

         - *filename* [default=None] the output filename, default is <model_name>.xml
         - *directory* [default=None] by default use filename otherwise join, <dir><filename>

        """
        if filename == None:
            filename = self.name + '.xml'
        if directory != None:
            assert os.path.exists(directory), '\n%s does not exist.' % directory
            filename = os.path.join(directory, filename)
        print('Writing file: %s' % filename)

        try:
            UseR = getuser()
        except:
            UseR = ''
        h1 = '<?xml version="1.0" encoding="utf-8"?>\n'
        h1 += (
            '<!-- Created with PySCeS ('
            + __version__
            + ') on '
            + time.strftime("%a, %d %b %Y %H:%M:%S")
            + ' by '
            + UseR
            + ' -->\n'
        )
        F = open(filename, 'w')
        F.write(h1 + self.getSBML())
        F.flush()
        F.close()
        print('Model %s exported as: %s' % (self.name, filename))


class SbmlToCore(object):
    SBML = None
    level = 2
    version = 5
    sbml_string = None
    sbml_file = None
    model = None
    document = None
    core = None
    # old
    ModelFile = None
    InitParams = None
    fixed_species = None
    species = None
    parameters = None
    reactions = None
    modifiers = None
    compartments = None
    _Function_time = None
    _Function_user = None
    _Function_init = None
    __nDict__ = None
    __functions__ = None
    __rules__ = None
    time = None
    __eDict__ = None
    notes = None
    IS_SINGLE_COMPARTMENT = False
    SPECIES_IN_AMOUNTS = False
    __reserved__ = None
    __sDict__ = None
    __InitDict__ = None
    __KeyWords__ = None
    __Errors__ = None
    __piecewises__ = None
    __error_sleep_time__ = 1
    _ecount = None
    __uDict__ = None
    __DEBUG__ = False
    __model_in_level__ = None
    __model_in_version__ = None
    COMP_FUDGE_FACTOR = 1.0

    def __init__(self):
        try:
            import libsbml as SBML

            self.SBML = SBML
        except Exception as e:
            print(e)
            print('SBML load error')
            self.SBML = None
        assert self.SBML != None, '\nNo SBML library available'

        self.__reserved__ = {'lambda': 'Lambda'}
        self.__KeyWords__ = {
            'Description': 'ModelDescription',
            'Modelname': 'PySCeSModel',
            'Species_In_Conc': True,
            'Output_In_Conc': True,
        }
        self.__piecewises__ = {}

    def sbmlFormulaToInfix(self, formula):
        """
        Convert and SBML MathML formula to infix. Returns an infix string representation of a formula

        - **formula** SBML AST math

        """
        infix = self.SBML.formulaToL3String(formula)
        infix = infix.replace('^', '**').replace('&&', 'and').replace('||','or')
        return infix


    def searchForCsymbolTime(self, sbml_math):
        """
        Takes SBML math xxx.getMath() and returns the results of a regex multi search for the CSymbol time
        r'<csymbol encoding="text" definitionURL="http://www.sbml.org/sbml/symbols/time">.*<'

        - **sbml_math** SBML math generated with math_element.getMath()

        """

        # check for csymbol time
        csymb = r'<csymbol encoding="text" definitionURL="http://www.sbml.org/sbml/symbols/time">.*<'
        hasTimeS = re.search(
            csymb, self.SBML.writeMathMLToString(sbml_math)
        )

        return hasTimeS


    def setReservedTerm(self, term, replacement):
        self.__reserved__.update({term: replacement})

    def getSbmlStringFromDisk(self, sbml, Dir=None):
        if sbml[-4:] != '.xml':
            print("Assuming .xml extension")
            sbml += '.xml'
        if Dir is None:
            Dir = CWD
        assert os.path.exists(
            os.path.join(Dir, sbml)
        ), '\nFile %s does not exist' % os.path.join(Dir, sbml)
        self.sbml_file = os.path.join(Dir, sbml)
        sbmlF = open(self.sbml_file, 'r')
        self.sbml_string = sbmlF.read()
        sbmlF.close()

    def getSbmlStringFromString(self, sbml_string):
        self.sbml_string = sbml_string
        self.sbml_file = os.path.join(CWD, 'sbml_string_loader.xml')

    def getSbmlModel(self, document=None):
        r = self.SBML.SBMLReader()
        if document is None:
            self.document = r.readSBMLFromString(self.sbml_string)
        else:
            self.document = document
            self.sbml_string = self.document.toSBML()
            self.sbml_file = os.path.join(CWD, 'sbml_string_loader.xml')
        self.model = self.document.getModel()
        self.__model_in_level__ = self.model.getLevel()
        self.__model_in_version__ = self.model.getVersion()
        if self.model.getId() not in ['untitled','']:
            self.__KeyWords__['Modelname'] = self.model.getId()
        else:
            self.__KeyWords__['Modelname'] = self.model.getName()
        self.__KeyWords__['Description'] = self.model.getName()

    def checkSbmlSupport(self, action=2):
        """
        Check level and warn about unsupported functionality, errors logged to sbml_conversion_errors.txt

        - **action** [default=2] action to take: 1, raise error, 2 print warning with delay, 3 silently ignore

        """
        print('Check SBML support is at action level {}'.format(2))
        print('SBML file is L{}V{}'.format(self.__model_in_level__, self.__model_in_version__))
        logfile = '{}-sbml_conversion_errors.txt'.format(os.path.split(self.sbml_file)[-1])
        warn = []
        fatal = []

        if self.model.getLevel() >  2:
            warn.append('Model is encoded as SBML Level 3, PySCeS only officially supports L2V5.')
        if self.model.getNumConstraints() > 0:
            fatal.append('PySCeS does not support Constraints.')
        if self.model.getNumInitialAssignments() > 0:
            fatal.append('PySCeS does not support InitialAssignments.')
        if self.model.getNumEvents() > 0 and self.model.getLevel() >  2:
            fatal.append('PySCeS does not support L3 events.')
        if len([r.getType() for r in  self.model.getListOfRules() if r.getType() > 1]) > 0:
            fatal.append('PySCeS only supports rate and assignment rules.')


        if len(warn) == 0 and len(fatal) == 0:
            return
        else:
            print('\n\n\n*********ERRORS***********\n')
            with open(logfile, 'w') as F:
                for err in warn:
                    print('WARNING: {}'.format(err))
                    F.write('WARNING: {}\n'.format(err))
                for err in fatal:
                    print('FATAL: {}'.format(err))
                    F.write('FATAL: {}\n'.format(err))
            print('\n*********ERRORS***********\n')
            print('\nPossible errors detected in SBML conversion, Model may be incomplete. Please check the error log file \"{}\" for details.\n'.format(logfile))
            if action == 1:
                raise RuntimeError
            elif action == 2:
                time.sleep(5)
            else:
                time.sleep(0.1)


    def getUnits(self):
        self.__uDict__ = {
            'substance': {'exponent': 1, 'multiplier': 1.0, 'scale': 0, 'kind': 'mole'},
            'volume': {'exponent': 1, 'multiplier': 1.0, 'scale': 0, 'kind': 'litre'},
            'time': {'exponent': 1, 'multiplier': 1.0, 'scale': 0, 'kind': 'second'},
            'length': {'exponent': 1, 'multiplier': 1.0, 'scale': 0, 'kind': 'metre'},
            'area': {'exponent': 2, 'multiplier': 1.0, 'scale': 0, 'kind': 'metre'},
        }
        for ud in self.model.getListOfUnitDefinitions():
            self.__uDict__.update({ud.getId(): {}})
            if ud.getNumUnits() == 1:
                u = ud.getListOfUnits()[0]
                self.__uDict__[ud.getId()].update(
                    {
                        'multiplier': u.getMultiplier(),
                        'exponent': u.getExponent(),
                        'scale': u.getScale(),
                        'kind': self.SBML.UnitKind_toString(u.getKind()),
                    }
                )
            else:
                for u in ud.getListOfUnits():
                    self.__uDict__[ud.getId()].update(
                        {
                            self.SBML.UnitKind_toString(u.getKind()): {
                                'multiplier': u.getMultiplier(),
                                'exponent': u.getExponent(),
                                'scale': u.getScale(),
                                'kind': self.SBML.UnitKind_toString(u.getKind()),
                            }
                        }
                    )

    def getEvents(self):
        self.__eDict__ = {}
        for ev in self.model.getListOfEvents():
            name = self.getId(ev)
            trigger = ev.getTrigger()
            triggerf = self.sbmlFormulaToInfix(trigger.getMath())

            # check for csymbol time
            hasTimeS = self.searchForCsymbolTime(trigger.getMath())
            tSymb = None
            if hasTimeS != None:
                tSymb = hasTimeS.group()[hasTimeS.group().find('>') :]
                tSymb = tSymb.replace('>', '').replace('<', '').strip()
                print(
                    'csymbol time defined as \"%s\" in event %s'
                    % (tSymb, trigger.getId())
                )

            delay = ev.getDelay()
            if delay != None:
                delay = self.sbmlFormulaToInfix(ev.getDelay().getMath())
            else:
                delay = 0.0

            self.__eDict__.update(
                {
                    name: {
                        'name': name,
                        'trigger': triggerf,
                        'delay': delay,
                        'assignments': {},
                        'tsymb': tSymb,
                    }
                }
            )
            for a in ev.getListOfEventAssignments():
                self.__eDict__[name]['assignments'].update(
                    {a.getVariable(): self.sbmlFormulaToInfix(a.getMath())}
                )

    def checkParsedInfix(self):
        assert len(InfixParser.SymbolErrors) == 0, (
            '\nUndefined symbols:\n%s' % InfixParser.SymbolErrors
        )
        assert InfixParser.LexOK, '\nLexer Failure:\n%s' % InfixParser.LexErrors
        assert InfixParser.ParseOK, '\nParser Failure:\n%s' % InfixParser.ParseErrors

    def getId(self, e):
        name = e.getId()
        if name in self.__reserved__:
            print(
                '%s is a reserved symbol, replacing with %s'
                % (name, self.__reserved__[name])
            )
            self.__Errors__.update(
                {
                    next(self._ecount): 'Reserved symbol %s replaced with %s'
                    % (name, self.__reserved__[name])
                }
            )
            name = self.__reserved__[name]
        return name

    def updatePiecewiseDict(self, piecewises):
        if len(list(piecewises.keys())) > 0:
            for p in piecewises:
                if len(list(piecewises[p].keys())) == 2:
                    piecewises[p][0].reverse()
                self.__piecewises__.update({p: piecewises[p]})

    def getParsedModel(self):
        init_fixed = {}
        init_var = {}
        init_par = {}
        self.__compartments__ = {}
        # new stuff
        self.__functions__ = {}
        self.SPECIES_IN_AMOUNTS = False
        self.__sDict__ = {}
        self.__InitDict__ = {}
        self.__Errors__ = {}
        self._ecount = itertools.count()

        # Initialise compartment volumes as a parameter - brett 20050908
        for comp in self.model.getListOfCompartments():
            cont = comp.getOutside()
            if cont == '':
                cont = None
            cSize = float(comp.getSize())
            self.__compartments__.update(
                {
                    self.getId(comp): {
                        'name': self.getId(comp),
                        ##  'volume': float(comp.getVolume()),
                        'size': cSize,
                        'dimensions': int(comp.getSpatialDimensions()),
                        'compartment': cont,
                        'area': None,
                        'fullName': comp.getName(),
                    }
                }
            )
        self.COMP_FUDGE_FACTOR = 1.0
        startFudging = 1.0e-6
        I_AM_FUDGING = False
        tmp = min(
            [
                abs(self.__compartments__[c]['size'])
                for c in list(self.__compartments__.keys())
            ]
        )
        if tmp < startFudging:
            self.COMP_FUDGE_FACTOR = tmp  # sneaky b%*))_)%$^&*@rds

        for c in list(self.__compartments__.keys()):
            if self.COMP_FUDGE_FACTOR < startFudging:
                newsize = self.__compartments__[c]['size'] / self.COMP_FUDGE_FACTOR
                print(
                    'INFO: Rescaling compartment with size %s to %s'
                    % (self.__compartments__[c]['size'], newsize)
                )
                self.__compartments__[c]['size'] = newsize
                self.__compartments__[c].update({'scale': self.COMP_FUDGE_FACTOR})
                I_AM_FUDGING = True

        if len(self.__compartments__) == 1:
            self.IS_SINGLE_COMPARTMENT = True
        else:
            self.IS_SINGLE_COMPARTMENT = False

        if len(self.model.getListOfSpecies()) < 1:
            self.__Errors__.update(
                {next(self._ecount): 'No free species!? ... help I\'m confused!'}
            )
        for i in self.model.getListOfSpecies():
            specname = self.getId(i)
            if i.getHasOnlySubstanceUnits():
                if not self.SPECIES_IN_AMOUNTS:
                    self.SPECIES_IN_AMOUNTS = True
        if self.SPECIES_IN_AMOUNTS:
            self.__KeyWords__['Species_In_Conc'] = False
        else:
            self.__KeyWords__['Species_In_Conc'] = True
            self.__KeyWords__['Output_In_Conc'] = True

        for i in self.model.getListOfSpecies():
            specname = self.getId(i)
            if self.__DEBUG__:
                print(
                    '%s has only substance units: %s'
                    % (specname, i.getHasOnlySubstanceUnits())
                )
            IS_CONC = i.isSetInitialConcentration()
            IS_AMNT = i.isSetInitialAmount()
            Conc = float(i.getInitialConcentration())
            Amount = float(i.getInitialAmount())
            if self.SPECIES_IN_AMOUNTS or IS_AMNT:
                if I_AM_FUDGING:
                    Conc = Amount / self.COMP_FUDGE_FACTOR
                else:
                    Conc = Amount
            fxd = None
            if i.getBoundaryCondition() or i.getConstant():
                if i.getConstant() and not i.getBoundaryCondition():
                    print(
                        '%s is set as constant, assuming: BoundaryCondition = True'
                        % specname
                    )
                init_fixed.setdefault(specname, Conc)
                fxd = True
            else:
                init_var.setdefault(specname, Conc)
                fxd = False

            self.__sDict__.update(
                {
                    specname: {
                        'name': specname,
                        'initial': Conc,
                        'compartment': i.getCompartment(),
                        'fixed': fxd,
                        'isamount': i.getHasOnlySubstanceUnits(),
                        'fullName': i.getName(),
                    }
                }
            )

        reactions = self.model.getListOfReactions()

        NetworkDict = dict(
            [
                (
                    i,
                    dict.fromkeys(
                        [
                            'Params',
                            'RateEq',
                            'Reagents',
                            'AllReagents',
                            'Modifiers',
                            'Type',
                        ]
                    ),
                )
                for i in [self.getId(j) for j in reactions]
            ]
        )

        # Add global parameters
        Punits = []

        if len(self.model.getListOfParameters()) > 0:
            for xp in self.model.getListOfParameters():
                if xp.isSetUnits():
                    Punits.append(self.getId(xp))
                init_par.update({self.getId(xp): float(xp.getValue())})
                # TODO: this needs to be overwritten when initial assignments are introduced
                if numpy.isnan(xp.getValue()):
                    print('ERROR: NaN detected as parameter value: {} setting to default value: 0.0'.format(self.getId(xp)))
                    #print(self.getId(xp), xp.getValue(), type(xp.getValue()))
                    init_par[self.getId(xp)] = 0.0
                    time.sleep(1)

        if len(Punits) > 0:
            self.__Errors__.update(
                {
                    next(self._ecount): 'Parameter units ignored for parameters:\n%s'
                    % Punits
                }
            )

        # add any function definitions
        if self.model.getNumFunctionDefinitions() > 0:
            for fnc in self.model.getListOfFunctionDefinitions():
                name = self.getId(fnc)

                func = self.sbmlFormulaToInfix(fnc.getMath())
                if 'lambda' in func:
                    func = func.replace('lambda', '')[1:-1]

                args = []
                for _ in range(fnc.getNumArguments()):
                    arg = func[: func.find(',')]
                    func = func[func.find(',') + 1 :]
                    args.append(arg)
                args = [s.strip() for s in args]
                func = func.strip()

                # check for csymbol time
                hasTimeS = self.searchForCsymbolTime(fnc.getMath())
                tSymb = None
                if hasTimeS != None:
                    tSymb = hasTimeS.group()[hasTimeS.group().find('>') :]
                    tSymb = tSymb.replace('>', '').replace('<', '').strip()
                    print(
                        'csymbol time defined as \"%s\" in formula %s'
                        % (tSymb, self.getId(fnc))
                    )
                    SRs = {tSymb: '_TIME_'}
                    SRs.update(self.__reserved__)
                    InfixParser.SymbolReplacements = SRs
                InfixParser.setNameStr('', '')
                InfixParser.parse(func)
                p_names = InfixParser.names
                func = InfixParser.output
                self.checkParsedInfix()
                # update piecewise dict ... don't ask!
                self.updatePiecewiseDict(InfixParser.piecewises)

                self.__functions__.update(
                    {name: {'args': args, 'formula': func, 'name': name}}
                )

        # ======= per reaction stuff ========
        hasFast = []
        Punits = []
        delayignore = []
        for i in reactions:
            rDict = NetworkDict[self.getId(i)]
            rDict.update({'name': self.getId(i)})
            rDict.update({'fullName': i.getName()})
            j = i.getKineticLaw()
            # ignore fast and warn
            if i.isSetFast():
                hasFast.append(self.getId(i))

            par = []
            if j != None:
                #req = j.getFormula()
                req = self.sbmlFormulaToInfix(j.getMath())
                p_names = None

                # check for csymbol time
                hasTimeS = self.searchForCsymbolTime(j.getMath())

                tSymb = None
                if hasTimeS != None:
                    tSymb = hasTimeS.group()[hasTimeS.group().find('>') :]
                    tSymb = tSymb.replace('>', '').replace('<', '').strip()
                    print(
                        'csymbol time defined as \"%s\" in reaction %s'
                        % (tSymb, self.getId(i))
                    )
                # if there are local parameters hash them to R_P
                if len(j.getListOfParameters()) > 0:
                    InfixParser.setNameStr('@', '@')
                    if hasTimeS != None:
                        SRs = {tSymb: '_TIME_'}
                        SRs.update(self.__reserved__)
                        InfixParser.SymbolReplacements = SRs
                    else:
                        InfixParser.SymbolReplacements = self.__reserved__
                    InfixParser.parse(req)
                    p_names = InfixParser.names
                    # catch delay and friends (ie check for shitty rate equations)
                    self.checkParsedInfix()
                    req = InfixParser.output
                    if InfixParser.DelayRemoved:
                        delayignore.append(self.getId(i))
                    if self.__DEBUG__:
                        print('Setting local parameter:')
                    for k in j.getListOfParameters():
                        if k.isSetUnits() and self.getId(k) not in Punits:
                            Punits.append(self.getId(k))
                        # TODO: replace with a regular expression based strategy
                        req = req.replace(
                            '@' + self.getId(k) + '@',
                            self.getId(i) + '_' + self.getId(k),
                        )
                        par.append(self.getId(i) + '_' + self.getId(k))
                        init_par.setdefault(
                            self.getId(i) + '_' + self.getId(k), k.getValue()
                        )
                        # local parameters are hashed with reaction name
                        if self.__DEBUG__:
                            print('%s_%s' % (self.getId(i), self.getId(k)), end=' ')
                    if self.__DEBUG__:
                        print('')
                    req = req.replace('@', '')
                else:
                    InfixParser.setNameStr('', '')
                    if hasTimeS != None:
                        SRs = {tSymb: '_TIME_'}
                        SRs.update(self.__reserved__)
                        InfixParser.SymbolReplacements = SRs
                    else:
                        InfixParser.SymbolReplacements = self.__reserved__
                    InfixParser.parse(req)
                    p_names = InfixParser.names
                    req = InfixParser.output
                    if InfixParser.DelayRemoved:
                        delayignore.append(self.getId(i))
                    self.checkParsedInfix()

                # update piecewise dict ... don't ask!
                self.updatePiecewiseDict(InfixParser.piecewises)
            else:
                print(
                    'WARNING: No rate equation for reaction %s in file: %s'
                    % (self.getId(i), self.sbml_file)
                )
                req = None

            rDict['Params'] = par
            rDict['RateEq'] = req
            rDict['compartment'] = None

            Substrates = []
            Products = []

            for k in i.getListOfReactants():
                species = k.getSpecies()
                stoich = -k.getStoichiometry()
                Substrates.append((species, stoich))
                # kill/collect stoichiometrymath for future processing
                smath = k.getStoichiometryMath()
                if smath != None:
                    self.__Errors__.update(
                        {
                            next(
                                self._ecount
                            ): 'StoichiometryMath (%s) not supported and ignored'
                            % k.getSpecies()
                        }
                    )

            for k in i.getListOfProducts():
                species = k.getSpecies()
                stoich = k.getStoichiometry()
                Products.append((species, stoich))
                # kill/collect stoichiometrymath for future processing
                smath = k.getStoichiometryMath()
                if smath != None:
                    self.__Errors__.update(
                        {
                            next(
                                self._ecount
                            ): 'StoichiometryMath (%s) not supported and ignored'
                            % k.getSpecies()
                        }
                    )

            # work with net stoichiometries
            rDict['Reagents'] = {}
            rDict['AllReagents'] = Substrates + Products
            rtmp = rDict['Reagents']

            for sp in Substrates + Products:
                if sp[0] not in rtmp:
                    rtmp.update({sp[0]: sp[1]})
                else:
                    rtmp[sp[0]] += sp[1]
            for r in list(rtmp.keys()):
                if abs(rtmp[r]) < 1.0e-14:
                    rtmp.pop(r)
            if i.getReversible() == True:
                t = 'Rever'
            else:
                t = 'Irrev'
            rDict['Type'] = t
            NetworkDict[self.getId(i)].update(rDict)

            mods = []
            if len(i.getListOfModifiers()) > 0:
                for m in i.getListOfModifiers():
                    # TODO: this should be done with SBML object.attributes 20110523
                    msbml = m.toSBML().split('species=\"')[1].split('\"')[0]
                    mods.append(msbml)
            else:
                mods = []
            rDict['Modifiers'] = mods
            # print "Modifiers", rDict['Modifiers']
        if len(hasFast) > 0:
            self.__Errors__.update(
                {
                    next(self._ecount): 'Fast attribute ignored for reactions:\n%s'
                    % hasFast
                }
            )
        if len(Punits) > 0:
            self.__Errors__.update(
                {
                    next(
                        self._ecount
                    ): 'Parameter units ignored for (local) parameters:\n%s'
                    % Punits
                }
            )
        if len(delayignore) > 0:
            self.__Errors__.update(
                {
                    next(self._ecount): 'delay function removed in reactions:\n%s'
                    % delayignore
                }
            )

        del hasFast, Punits, delayignore

        # ======= Add Other Model components ========
        self.getEvents()
        self.getRules()

        self.notes = ''
        if self.model.getNotes() is not None:
            self.notes = self.model.getNotesString().replace('\n', '\n# ')

        self.ModelFile = self.model.getId()
        if len(self.ModelFile) == 0:
            self.ModelFile = self.model.getName()

        InitStrings = []

        for s in list(init_var.keys()):
            self.__InitDict__.update({s: float(init_var[s])})
            setattr(self, s, float(init_var[s]))
        for s in list(init_fixed.keys()):
            self.__InitDict__.update({s: float(init_fixed[s])})
            setattr(self, s, float(init_fixed[s]))
        for s in list(init_par.keys()):
            self.__InitDict__.update({s: float(init_par[s])})
            setattr(self, s, float(init_par[s]))

        self.InitParams = list(init_par.keys())
        self.fixed_species = list(init_fixed.keys())
        self.species = list(init_var.keys())
        self.parameters = list(init_par.keys())
        self.reactions = list(NetworkDict.keys())
        self.modifiers = []
        for r in list(NetworkDict.keys()):
            self.modifiers.append((r, NetworkDict[r]['Modifiers']))
        if len(list(self.__rules__.keys())) > 0:
            self._Function_forced = ''
            for r in list(self.__rules__.keys()):
                self._Function_forced += (
                    self.__rules__[r]['name']
                    + ' = '
                    + self.__rules__[r]['formula']
                    + '\n'
                )
        else:
            self._Function_forced = 'pass\n'
        self._Function_time = ''
        self._Function_user = ''
        self._Function_init = ''
        self.__nDict__ = NetworkDict

        if len(list(self.__Errors__.keys())) > 0:
            print(
                '\n*******************************************************************'
            )
            ##  print 'Errors encountered in SBML translated!'
            print('Issues encountered in SBML translation (model processed anyway)')
            print('SBML source: %s' % self.sbml_file)
            print(
                '*******************************************************************\n'
            )
            ekeys = list(self.__Errors__.keys())
            ekeys.sort()
            for e in ekeys:
                print(self.__Errors__[e], '\n')
            print(
                '*******************************************************************\n'
            )
            time.sleep(self.__error_sleep_time__)

    def getRules(self):
        self.__rules__ = {}
        for rule in self.model.getListOfRules():
            rtype = 'unknown'
            if rule.isAssignment():
                rtype = 'assignment'
            elif rule.isRate():
                rtype = 'rate'
            elif rule.isAlgebraic():
                rtype = 'algebraic'
                self.__Errors__.update(
                    {
                        next(self._ecount): 'Algebraic rule (%s) ignored'
                        #% rule.getFormula()
                        % self.sbmlFormulaToInfix(rule.getMath())
                    }
                )

            assert rtype in ['assignment', 'rate', 'algebraic'], (
                '\n%s rules currently not supported.' % rtype
            )

            # check for csymbol time
            hasTimeS = self.searchForCsymbolTime(rule.getMath())

            tSymb = None
            #formula = rule.getFormula()
            formula = self.sbmlFormulaToInfix(rule.getMath())
            p_names = None
            if hasTimeS != None:
                tSymb = hasTimeS.group()[hasTimeS.group().find('>') :]
                tSymb = tSymb.replace('>', '').replace('<', '').strip()
                print(
                    'csymbol time defined as \"%s\" in rule %s'
                    % (tSymb, self.getId(rule))
                )
                InfixParser.setNameStr('', '')
                SRs = {tSymb: '_TIME_'}
                SRs.update(self.__reserved__)
                InfixParser.SymbolReplacements = SRs
                InfixParser.parse(formula)
                p_names = InfixParser.names
                formula = InfixParser.output
                pws = InfixParser.piecewises
                self.checkParsedInfix()
            else:
                InfixParser.setNameStr('', '')
                InfixParser.SymbolReplacements = self.__reserved__
                InfixParser.parse(formula)
                p_names = InfixParser.names
                formula = InfixParser.output
                pws = InfixParser.piecewises
                self.checkParsedInfix()

            # update piecewise dict ... don't ask!
            self.updatePiecewiseDict(InfixParser.piecewises)

            self.__rules__.update(
                {
                    rule.getVariable(): {
                        'name': rule.getVariable(),
                        'formula': formula,
                        'type': rtype,
                        '_names': p_names,
                    }
                }
            )

    def removeCompartmentFromKineticLaw(self, kl):
        strBuf = io.StringIO()
        mathMLin = self.sbmlFormulaToInfix(kl.getMath())
        strBuf.write(mathMLin)
        strBuf.seek(0)

        etree = ElementTree.parse(strBuf)
        root = etree.getroot()

        node_idx = {}
        counter = itertools.count(1)
        self._comp_idx = []

        def idxNode(node, idx=0):
            ##  print 'node.tag', node.tag
            node.attrib.update({'idx': str(idx)})
            node_idx.update({idx: node})
            if node.text != None and node.text.strip() in list(
                self.__compartments__.keys()
            ):
                node.attrib.update({'compartment': '1'})
                self._comp_idx.append(idx)
            else:
                node.attrib.update({'compartment': '0'})

            children = node.getchildren()
            for child in range(len(children)):
                children[child].attrib.update({'parent': str(idx)})
                ##  print '\tchild.tag', children[child].tag
                idxNode(children[child], next(counter))

        idxNode(root, idx=0)

        if len(self._comp_idx) > 0:
            for comp_idx in self._comp_idx:
                cParent = int(node_idx[comp_idx].attrib['parent'])
                if (
                    len(node_idx[cParent].getchildren()) == 3
                    and node_idx[cParent].getchildren()[0].tag
                    == '{http://www.w3.org/1998/Math/MathML}times'
                ):
                    for c in node_idx[cParent].getchildren():
                        if c.tag != '{http://www.w3.org/1998/Math/MathML}times' and c.text.strip() not in list(
                            self.__compartments__.keys()
                        ):
                            cParentParent = int(node_idx[cParent].attrib['parent'])
                            for cp in node_idx[cParentParent].getchildren():
                                if cp.attrib['idx'] == node_idx[cParent].attrib['idx']:
                                    cidx = node_idx[cParentParent]._children.index(cp)
                                    node_idx[cParentParent].remove(node_idx[cParent])
                                    c.attrib.update(
                                        {
                                            'parent': node_idx[cParentParent].attrib[
                                                'idx'
                                            ]
                                        }
                                    )
                                    node_idx[cParentParent].insert(cidx, c)
                                    break
                elif (
                    len(node_idx[cParent].getchildren()) == 2
                    and node_idx[cParent].getchildren()[0].tag
                    == '{http://www.w3.org/1998/Math/MathML}minus'
                ):
                    for c in node_idx[cParent].getchildren():
                        ##  print c.tag, c.attrib
                        if c.tag == '{http://www.w3.org/1998/Math/MathML}minus':
                            cParentParent = int(node_idx[cParent].attrib['parent'])
                            for cp in node_idx[cParentParent].getchildren():
                                ##  print cp.attrib['idx'], node_idx[cParent].attrib['idx']
                                if cp.attrib['idx'] == node_idx[cParent].attrib['idx']:
                                    cidx = node_idx[cParentParent]._children.index(cp)
                                    node_idx[cParentParent].remove(node_idx[cParent])
                                    c.attrib.update(
                                        {
                                            'parent': node_idx[cParentParent].attrib[
                                                'idx'
                                            ]
                                        }
                                    )
                                    node_idx[cParentParent].insert(cidx, c)
                                    break
                elif len(node_idx[cParent].getchildren()) > 3:
                    for c in node_idx[cParent].getchildren():
                        if c.text != None and c.text.strip() in list(
                            self.__compartments__.keys()
                        ):
                            node_idx[cParent].remove(node_idx[comp_idx])
                            break
        for n in node_idx:
            if 'idx' in node_idx[n].attrib:
                node_idx[n].attrib.pop('idx')
            if 'compartment' in node_idx[n].attrib:
                node_idx[n].attrib.pop('compartment')
            if 'parent' in node_idx[n].attrib:
                node_idx[n].attrib.pop('parent')

        strBuf = io.StringIO()
        etree.write(strBuf)
        strBuf.seek(0)
        mathMLout = strBuf.read()
        newAST = self.SBML.readMathMLFromString(mathMLout)
        kl.setMath(newAST)



import pysces
try:
    input = raw_input  # Py2 compatibility
except NameError:
    pass

# Creates a ratechar object for the lin5_hill model
rc = pysces.RateChar('lin5_hill')

# This function does parameter scans for each species in the patway
rc.doAllRateChar()

# getRateCharData returns the ratechar data for a certain species (C in
# this case). It has some legacy functions and a function to write the 
# data to file and a function to find determine if the species is 
# regulatory for any reactions
rcdC = rc.getRateCharData('C')


print('\n\n\n')
print('C is regulatory for the following reactions:')
# This loop prints each reaction for which C is a regulatory metabolite
for reaction in rcdC.findRegulatory():
    print(reaction)
    print('\n')
    


input('Press Enter to continue...')


# The figure function returns a figure object that is used to plot the 
# data stored in the ratechardata object
figC = rcdC.figure()


# The functions below create specific plots from the data. The 'indiv' 
# functions are useful when there are multiple supply or demand reactions
# as they plot each reaction on it's own figure. 
figC.plotRateChar()
input('Press Enter to continue...')
figC.plotElasRC()
input('Press Enter to continue...')
figC.plotElasRCIndiv()
input('Press Enter to continue...')
figC.plotPartialRCDemand()
input('Press Enter to continue...')
figC.plotPartialRCSupply()
input('Press Enter to continue...')

# The figure object is even more powerful as it also allows you to create
# custom plots. 

# individual lines can  be switched on or off using the set_visible method
# use the reactions names as set out in the reactions dictionary (fig.reactions 
# in this example)
# e for elasticity
# f for flux
# r for response coefficient
# p for partial response coefficients
figC._clear()
# clears the figure from any previous plots

figC.title_label = 'Model: lin5_hill - Green line intersecting a red line'
# changing the value of this variable sets the title of the figure

figC.set_visible('J_R3',f=True)
figC.set_visible('J_R4',f=True)

figC.show()
#shows the figure

figC.save('testfigure1.png')  
#saves the figure with the filename provided in the pysces directory

input('Press ENTER to continue...')



# set_type_visible sets the visibility of all lines of a certain type
# The options are:
#    supply_flux
#    demand_flux
#    supply_rc
#    demand_rc
#    supply_elas
#    demand_elas
#    mod_elas
#    supply_partial_rc
#    demand_partial_rc
figC._clear()
figC.title_label = ''
figC.set_type_visible('supply_flux',True)
figC.set_type_visible('demand_flux',True)
figC.set_type_visible('supply_partial_rc',True,affzero=True) 
## when affzero is true, partial response coefficients with slopes with
## a gradient of zero will also be affected 
figC.show()
figC.save('testfigure2.png') 
input('Press ENTER to continue...')




# In addition to showing the default plots on the screen, we can also save
# them in the ratechar_out directory under the name of the model and with
# subdirectories for each species.
for species in rc._psc_mod.species:
    rcdX = rc.getRateCharData(species)
    figX = rcdX.figure()
    
    figX.plotRateChar(1)
    figX.plotElasRC(1)
    figX.plotPartialRCSupply(1)
    figX.plotPartialRCDemand(1)
    
    







import os

os.chdir('c:/mypysces/kraken')
##  os.chdir('/home/bgoli/mypysces/kraken')
from pysces.kraken.Kraken import *
## from Kraken import *

# build controller
blob = getBasicController()
Model_File = '2006_isola2.psc'
Model_Dir = 'c:\\mypysces\\pscmodels\\isola06'
##  Model_Dir = '/home/bgoli/mypysces/pscmodels/isola06'
blob.startModelServer(Model_File, Model_Dir)
blob.startTentacleMonitor(interval=60)

# initialise tentacles
print('Server list: %s\n' % blob.available_server_list)
server_list = blob.available_server_list

I_list = ['P_INIT,' + blob.model_server.model_file_name, 'P_LOAD',\
    'P_SET_FLOAT,mode_state_init,3', 'P_SET_FLOAT,nleq2_nitmax,20',\
    'P_SET_FLOAT,mode_state_init3_factor,0.5',\
    'P_SET_FLOAT,V2,200.0', 'P_SET_FLOAT,V3,200.0']


# continuation
yrange = 100
xpoints = 512
delta_list = scipy.logspace(scipy.log10(0.2),scipy.log10(1.0), yrange)
print(delta_list)
delta_gen = buildCycler(delta_list)

def deltaCommand():
    delta = 'P_SET_FLOAT,A05,%s' % next(delta_gen)
    return [delta]

job_list = []
for d in delta_list:
    ##  job_list.append('P_CONTINUATION,P,0.001,4.0e3,%s,%s' % (xpoints,d))
    job_list.append('P_CONTINUATION,P,0.01,3.0e4,%s,%s' % (xpoints,d))


original_job_list = tuple(job_list)

TIME_START = time.time()
DaResults = []
JobsWaiting = True
getServer = buildCycler(blob.available_server_list)
while JobsWaiting:
    deferred_jobs = []
    job_servers = []
    for job in range(len(job_list)):
        if job < len(server_list):
            server = next(getServer)
            job_servers.append(server)
            blob.sendCmdListToOne(I_list, server)
            blob.sendCmdListToOne(deltaCommand(), server)
            blob.sendJobToOne(job_list[job], server)
        else:
            deferred_jobs.append(job_list[job])
            print('Job %s of %s deferred' % (job+1, len(job_list)))
    blob.runAllJobs()
    ##  blob.sendCmdListToAll(['P_STORE_DATA'])
    for server in job_servers:
        blob.getDataFromOneJob(server)
        arr = blob.tentacle_response[server]
        DaResults.append(arr)
        print(type(arr))
    blob.clearProcessMap()
    if len(deferred_jobs) > 0:
        print('\nDeferred:', deferred_jobs)
        job_list = deferred_jobs
    else:
        JobsWaiting = False
    cPickle.dump(DaResults, open('cont_result_list.bin','wb'), 2)


print('\nNumber of results = %s' % len(DaResults))
total_states = 0
for x in range(len(DaResults)):
    try:
        print('\tResult %s has %s rows and %s columns' % (x+1, DaResults[x].shape[0], DaResults[x].shape[1]))
        total_states += DaResults[x].shape[0]
    except:
        print(type(DaResults[x]))
        print('\tResult %s %s: %s' % (x+1, type(DaResults[x]), DaResults[x]))
        print(original_job_list[x])

TIME_END = time.time()

print('Total time taken to complete %s state scan %s minutes' % (total_states, (TIME_END-TIME_START)/60.0))

FinalArray = concatenateArrays(DaResults)
cPickle.dump(FinalArray, open('cont_result_array.bin','wb'), 2)


if __name__ == '__main__':
    import scipy.io

    # get rid of negative values in axes and fluxes
    FinalArrayClean = []
    FinalVTK = []
    for r in range(FinalArray.shape[0]):
        if FinalArray[r,0] <= 0.0 or FinalArray[r,1] <= 0.0 or FinalArray[r,4] <= 0.0:
            FinalArray[r,:] == 0.0
        else:
            FinalArrayClean.append(FinalArray[r,:])

    FinalArrayClean = numpy.array(FinalArrayClean)

    # write out an xyz gplt file
    coords = numpy.transpose(numpy.array((FinalArrayClean[:,0],FinalArrayClean[:,1],FinalArrayClean[:,4])))
    scipy.io.write_array('cont_result_gplt.dat', coords, separator=' ')

    # write out a f%$^&* VTK file

    """
    Orgiginal version http://mayavi.sourceforge.net/mwiki/PointClouds
    Hacked by brett 20070221

    Converts xyz point cloud into something renderable
    Original code by Rod Holland & Prabhu Ramachandran
    Handy to visualise raw ascii column file
    """

    ##  i=0
    ##  x=[]
    ##  y=[]
    ##  z=[]

    ##  # add log(data)
    ##  for r in range(coords.shape[0]):
        ##  x.append(scipy.log10(coords[r,0]))
        ##  y.append(scipy.log10(coords[r,1]))
        ##  z.append(scipy.log10(coords[r,2]))

    n=coords.shape[0]
    print("n:",n)
    # write data to vtk polydata file
    # write header
    out = open('cont_result_ug.vtk', 'w')
    h1 = "# vtk DataFile Version 2.0\n"
    h1 += "loop\n"
    h1 += "ASCII\n"
    h1 += "DATASET UNSTRUCTURED_GRID\n"
    h1 += "POINTS " + str(n) + " double\n"
    out.write(h1)
    # write xyz data
    for r in range(n):
        #s = '%15.2f %15.2f %15.2f' % (x[i], y[i], z[i])
        out.write(str(scipy.log10(coords[r,0]))+" "+str(scipy.log10(coords[r,1]))+" "+str(scipy.log10(coords[r,2]))+'\n')
        ##  out.write(str(scipy.log10(coords[r,2]))+" "+str(scipy.log10(coords[r,1]))+" "+str(scipy.log10(coords[r,0]))+'\n')

    # write cell data
    out.write("CELLS "+ str(n)+ " "+ str(2*n)+'\n')
    for r in range(n):
            #s = '1 %d \n' % (i)
            out.write("1 "+str(r)+"\n")

    # write cell types
    out.write("CELL_TYPES " + str(n)+'\n')
    for r in range(n):
        out.write("1 \n")

    # write z scalar values
    h2 = '\n' + """POINT_DATA """ + str(n) + "\n"
    h3 = "SCALARS Z_Value double 1\n"
    h3 += "LOOKUP_TABLE default\n"
    out.write(h2+ h3)

    for r in range(n):
        sc=(scipy.log10(coords[r,2]))
        out.write(str(sc)+ "\n")

    out.write('\n')
    out.close()


    #archived shit
    """
    ##  undup = {}
    ##  rounding = 5
    ##  for r in range(coords.shape[0]):
        ##  undup.setdefault(round(coords[r,2], rounding), [coords[r,0], coords[r,1]])
    ##  rF = open('cont_result_array_gplt_reduced.dat','w')
    ##  print 'Total states: %s\nReduced states: %s ' % (total_states, len(undup.keys()))
    ##  for key in undup:
        ##  rF.write('%s, %s, %s\n' % (undup[key][0], undup[key][1], key))
    ##  rF.flush()
    ##  rF.close()

    """
